# -*- coding: utf-8 -*-
import numpy as np
from ._stft import stft, get_window, _check_NOLA
from ._ssq_cwt import _invert_components, _process_component_inversion_args
from .utils import WARN, EPS, pi, _process_fs_and_t
from .ssqueezing import ssqueeze, _check_ssqueezing_args


def ssq_stft(x, window=None, n_fft=None, win_len=None, hop_len=1, fs=None, t=None,
             modulated=True, ssq_freqs=None, padtype='reflect', squeezing='sum',
             gamma=None, preserve_transform=True):
    """Synchrosqueezed Short-Time Fourier Transform.
    Implements the algorithm described in Sec. III of [1].

    MATLAB docs: https://www.mathworks.com/help/signal/ref/fsst.html

    # Arguments:
        x: np.ndarray
            Input vector, 1D.

        window, n_fft, win_len, hop_len, fs, t, padtype, modulated
            See `help(stft)`.

        ssq_freqs, squeezing
            See `help(ssqueezing.ssqueeze)`.

        gamma: float / None
            STFT phase threshold. Sets `w=inf` for small values of `Sx` where
            phase computation is unstable and inaccurate (like in DFT):
                w[abs(Sx) < beta] = inf
            This is used to zero `Sx` where `w=0` in computing `Tx` to ignore
            contributions from points with indeterminate phase.
            Default = sqrt(machine epsilon) = np.sqrt(np.finfo(np.float64).eps)

        preserve_transform: bool (default True)
            Whether to return `Sx` as directly output from `stft` (it might be
            altered by `ssqueeze` or `phase_transform`). Uses more memory
            per storing extra copy of `Sx`.

    # Returns:
        Tx: np.ndarray
            Synchrosqueezed STFT of `x`, of same shape as `Sx`.
        Sx: np.ndarray
            STFT of `x`. See `help(stft)`.
        ssq_freqs: np.ndarray
            Frequencies associated with rows of `Tx`.
        Sfs: np.ndarray
            Frequencies associated with rows of `Sx` (by default == `ssq_freqs`).
        w: np.ndarray
            Phase transform of STFT of `x`. See `help(phase_stft)`.
        dSx: np.ndarray
            Time-derivative of STFT of `x`. See `help(stft)`.

    # References:
        1. Synchrosqueezing-based Recovery of Instantaneous Frequency from
        Nonuniform Samples. G. Thakur and H.-T. Wu.
        https://arxiv.org/abs/1006.2533
    """
    _, fs, _ = _process_fs_and_t(fs, t, len(x))
    n_fft = n_fft or len(x)
    _check_ssqueezing_args(squeezing)

    Sx, dSx = stft(x, window, n_fft=n_fft, win_len=win_len, hop_len=hop_len,
                   fs=fs, padtype=padtype, modulated=modulated, derivative=True)
    _Sx = Sx.copy() if preserve_transform else Sx

    Sfs = np.linspace(0, .5, Sx.shape[0]) * fs
    w = phase_stft(_Sx, dSx, Sfs, gamma)

    if ssq_freqs is None:
        ssq_freqs = Sfs
    Tx, ssq_freqs = ssqueeze(_Sx, w, transform='stft', squeezing=squeezing,
                             ssq_freqs=ssq_freqs)

    return Tx, Sx, ssq_freqs, Sfs, w, dSx


def issq_stft(Tx, window=None, cc=None, cw=None, n_fft=None, win_len=None,
              hop_len=1, modulated=True):
    """Inverse synchrosqueezed STFT.

    # Arguments:
        x: np.ndarray
            Input vector, 1D.

        window, n_fft, win_len, hop_len, modulated
            See `help(stft)`. Must match those used in `ssq_stft`.

        cc, cw: np.ndarray
            See `help(issq_cwt)`.

    # Returns:
        x: np.ndarray
            Signal as reconstructed from `Tx`.

    # References:
        1. Synchrosqueezing-based Recovery of Instantaneous Frequency from
        Nonuniform Samples. G. Thakur and H.-T. Wu.
        https://arxiv.org/abs/1006.2533

        2. Fourier synchrosqueezed transform MATLAB docs.
        https://www.mathworks.com/help/signal/ref/fsst.html
    """
    def _process_args(Tx, window, cc, cw, win_len, hop_len, n_fft, modulated):
        if not modulated:
            raise ValueError("inversion with `modulated == False` "
                             "is unsupported.")
        if hop_len != 1:
            raise ValueError("inversion with `hop_len != 1` is unsupported.")

        cc, cw, full_inverse = _process_component_inversion_args(cc, cw)

        n_fft = n_fft or (Tx.shape[0] - 1) * 2
        win_len = win_len or n_fft

        window = get_window(window, win_len, n_fft=n_fft)
        _check_NOLA(window, hop_len)
        if abs(np.argmax(window) - len(window)//2) > 1:
            WARN("`window` maximum not centered; results may be inaccurate.")

        return window, cc, cw, win_len, hop_len, n_fft, full_inverse

    (window, cc, cw, win_len, hop_len, n_fft, full_inverse
     ) = _process_args(Tx, window, cc, cw, win_len, hop_len, n_fft, modulated)

    if full_inverse:
        # Integration over all frequencies recovers original signal
        x = Tx.real.sum(axis=0)
    else:
        x = _invert_components(Tx, cc, cw)

    x *= (2 / window[len(window)//2])
    return x


def phase_stft(Sx, dSx, Sfs, gamma=None):
    """Phase transform of STFT:
        w[u, k] = Im( k - d/dt(Sx[u, k]) / Sx[u, k] / (j*2pi) )

    Defined in Sec. 3 of [1]. Additionally explained in:
        https://dsp.stackexchange.com/a/72589/50076

    # Arguments:
        Sx: np.ndarray
            STFT of `x`, where `x` is 1D.

        dSx: np.ndarray
            Time-derivative of STFT of `x`

        Sfs: np.ndarray
            Associated physical frequencies, according to `dt` used in `stft`.
            Spans 0 to fs/2, linearly.

        gamma: float / None
            STFT phase threshold. Sets `w=inf` for small values of `Sx` where
            phase computation is unstable and inaccurate (like in DFT):
                w[abs(Sx) < beta] = inf
            This is used to zero `Wx` where `w=0` in computing `Tx` to ignore
            contributions from points with indeterminate phase.
            Default = sqrt(machine epsilon) = np.sqrt(np.finfo(np.float64).eps)

    # Returns:
        w: np.ndarray
            Phase transform for each element of `Sx`. w.shape == Sx.shape.

    # References:
        1. Synchrosqueezing-based Recovery of Instantaneous Frequency from
        Nonuniform Samples. G. Thakur and H.-T. Wu.
        https://arxiv.org/abs/1006.2533

        2. The Synchrosqueezing algorithm for time-varying spectral analysis:
        robustness properties and new paleoclimate applications.
        G. Thakur, E. Brevdo, N.-S. Fučkar, and H.-T. Wu.
        https://arxiv.org/abs/1105.0010
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        w = Sfs.reshape(-1, 1) - np.imag(dSx / Sx) / (2*pi)

    # treat negative phases as positive; these are in small minority, and
    # slightly aid invertibility (as less of `Wx` is zeroed in ssqueezing)
    w = np.abs(w)

    gamma = gamma or np.sqrt(EPS)
    # threshold out small points
    w[np.abs(Sx) < gamma] = np.inf
    return w



# -*- coding: utf-8 -*-
import numpy as np
from scipy import integrate
from .common import WARN, assert_is_one_of, p2up
from ..algos import _min_neglect_idx, find_maximum, find_first_occurrence

pi = np.pi

__all__ = [
    'adm_ssq',
    'adm_cwt',
    'cwt_scalebounds',
    'process_scales',
    'infer_scaletype',
    'make_scales',
    'logscale_transition_idx',
    'nv_from_scales',
    'find_min_scale',
    'find_max_scale',
    'find_downsampling_scale',
    'integrate_analytic',
    'find_max_scale_alt',
    '_process_fs_and_t',
]


def adm_ssq(wavelet):
    """Calculates the synchrosqueezing admissibility constant, the term
    R_psi in Eq 15 of [1] (also see Eq 2.5 of [2]). Uses numeric intergration.

        integral(conj(wavelet(w)) / w, w=0..inf)

    # References:
        1. The Synchrosqueezing algorithm for time-varying spectral analysis:
        robustness properties and new paleoclimate applications. G. Thakur,
        E. Brevdo, N.-S. Fučkar, and H.-T. Wu.
        https://arxiv.org/abs/1105.0010

        2. Synchrosqueezed Wavelet Transforms: a Tool for Empirical Mode
        Decomposition. I. Daubechies, J. Lu, H.T. Wu.
        https://arxiv.org/pdf/0912.2437.pdf
    """
    wavelet = Wavelet._init_if_not_isinstance(wavelet).fn
    Css = integrate_analytic(lambda w: np.conj(wavelet(w)) / w)
    Css = Css.real if abs(Css.imag) < 1e-15 else Css
    return Css


def adm_cwt(wavelet):
    """Calculates the cwt admissibility constant as per Eq. (4.67) of [1].
    Uses numeric integration.

        integral(wavelet(w) * conj(wavelet(w)) / w, w=0..inf)

    1. Wavelet Tour of Signal Processing, 3rd ed. S. Mallat.
    https://www.di.ens.fr/~mallat/papiers/WaveletTourChap1-2-3.pdf
    """
    wavelet = Wavelet._init_if_not_isinstance(wavelet).fn
    Cpsi = integrate_analytic(lambda w: np.conj(wavelet(w)) * wavelet(w) / w)
    Cpsi = Cpsi.real if abs(Cpsi.imag) < 1e-15 else Cpsi
    return Cpsi


def cwt_scalebounds(wavelet, N, preset=None, min_cutoff=None, max_cutoff=None,
                    cutoff=None, bin_loc=None, bin_amp=None, use_padded_N=True,
                    viz=False):
    """Finds range of scales for which `wavelet` is "well-behaved", as
    determined by `preset`.

    `min_scale`: found such that freq-domain wavelet takes on `cutoff` of its max
    value on the greatest bin.
      - Lesser `cutoff` -> lesser `min_scale`, always

    `max_scale`: search determined by `preset`:
        - 'maximal': found such that freq-domain takes `bin_amp` of its max value
          on the `bin_loc`-th (non-dc) bin
          - Greater `bin_loc` or lesser `bin_amp` -> lesser `max_scale`, always

        - 'minimal': found more intricately independent of precise bin location,
          but is likely to omit first several bins entirely; see
          `help(utils.find_max_scale_alt)`.
          - Greater `min_cutoff` -> lesser `max_scale`, generally

    `viz==2` for more visuals, `viz==3` for even more.

    # Arguments:
        wavelet: `wavelets.Wavelet`
            Wavelet sampled in Fourier frequency domain. See `help(cwt)`.

        N: int
            Length of wavelet to use.

        min_cutoff, max_cutoff: float > 0 / None
            Used to find max scale with `preset='minimal'`.
            See `help(utils.find_max_scale_alt)`

        cutoff: float / None
            Used to find min scale. See `help(utils.find_min_scale)`

        preset: str['maximal', 'minimal', 'naive'] / None
            - 'maximal': yields a larger max and smaller min.
            - 'minimal': strives to keep wavelet in "well-behaved" range of std_t
            and std_w, but very high or very low frequencies' energies will be
            under-represented. Is closer to MATLAB's default `cwtfreqbounds`.
            - 'naive': returns (1, N), which is per original MATLAB Toolbox,
            but a poor choice for most wavelet options.
            - None: will use `min_cutoff, max_cutoff, cutoff` values, else
            override `min_cutoff, max_cutoff` with those of `preset='minimal'`,
            and of `cutoff` with that of `preset='maximal'`:
                (min_cutoff, max_cutoff, cutoff) = (0.6, 0.8, -.5)

        use_padded_N: bool (default True)
            Whether to use `N=p2up(N)` in computations. Typically `N == len(x)`,
            but CWT pads to next power of 2, which is the actual wavelet length
            used, which typically behaves significantly differently at scale
            extrema, thus recommended default True. Differs from passing
            `N=p2up(N)[0]` and False only for first visual if `viz`, see code.

    # Returns:
        min_scale, max_scale: float, float
            Minimum & maximum scales.
    """
    def _process_args(preset, min_cutoff, max_cutoff, cutoff, bin_loc, bin_amp):
        defaults = dict(min_cutoff=.6, max_cutoff=.8, cutoff=-.5)

        if preset is not None:
            if any((min_cutoff, max_cutoff, cutoff)):
                WARN("`preset` will override `min_cutoff, max_cutoff, cutoff`")
            elif preset == 'minimal' and any((bin_amp, bin_loc)):
                WARN("`preset='minimal'` ignores `bin_amp` & `bin_loc`")

            assert_is_one_of(preset, 'preset',
                             ('maximal', 'minimal', 'naive'))
            if preset in ('naive', 'maximal'):
                min_cutoff, max_cutoff = None, None
                if preset == 'maximal':
                    cutoff = -.5
            else:
                min_cutoff, max_cutoff, cutoff = defaults.values()
        else:
            if min_cutoff is None:
                min_cutoff = defaults['min_cutoff']
            elif min_cutoff <= 0:
                raise ValueError("`min_cutoff` must be >0 (got %s)" % min_cutoff)

            if max_cutoff is None:
                max_cutoff = defaults['max_cutoff']
            elif max_cutoff < min_cutoff:
                raise ValueError("must have `max_cutoff > min_cutoff` "
                                 "(got %s, %s)" % (max_cutoff, min_cutoff))

        bin_loc = bin_loc or (1 if preset == 'maximal' else None)
        bin_amp = bin_amp or (1 if preset == 'maximal' else None)
        cutoff  = cutoff if (cutoff is not None) else defaults['cutoff']

        return min_cutoff, max_cutoff, cutoff, bin_loc, bin_amp

    def _viz():
        _viz_cwt_scalebounds(wavelet, N=M, Nt=M, min_scale=min_scale,
                             max_scale=max_scale, cutoff=cutoff)
        if viz >= 2:
            wavelet_waveforms(wavelet, M, min_scale)
            wavelet_waveforms(wavelet, M, max_scale)
        if viz == 3:
            scales = make_scales(M, min_scale, max_scale)
            sweep_harea(wavelet, M, scales)

    min_cutoff, max_cutoff, cutoff, bin_loc, bin_amp = _process_args(
        preset, min_cutoff, max_cutoff, cutoff, bin_loc, bin_amp)

    if preset == 'naive':  # still _process_args for the NOTE
        return 1, N

    M = p2up(N)[0] if use_padded_N else N
    min_scale = find_min_scale(wavelet, cutoff=cutoff)

    if preset in ('minimal', None):
        max_scale = find_max_scale_alt(wavelet, M, min_cutoff=min_cutoff,
                                       max_cutoff=max_cutoff)
    elif preset == 'maximal':
        max_scale = find_max_scale(wavelet, M, bin_loc=bin_loc, bin_amp=bin_amp)

    if viz:
        _viz()
    return min_scale, max_scale


def _assert_positive_integer(g, name=''):
    if not (g > 0 and float(g).is_integer()):
        raise ValueError(f"'{name}' must be a positive integer (got {g})")


def process_scales(scales, N, wavelet=None, nv=None, get_params=False,
                   use_padded_N=True):
    """Makes scales if `scales` is a string, else validates the array,
    and returns relevant parameters if requested.

        - Ensures, if array,  `scales` is 1D, or 2D with last dim == 1
        - Ensures, if string, `scales` is one of ('log', 'linear')
        - If `get_params`, also returns (`scaletype`, `nv`, `na`)
           - `scaletype`: inferred from `scales` ('linear' or 'log') if array
           - `nv`, `na`: computed newly only if not already passed
    """
    def _process_args(scales, nv, wavelet):
        preset = None
        if isinstance(scales, str):
            if ':' in scales:
                scales, preset = scales.split(':')
            elif scales == 'log-piecewise':
                preset = 'maximal'

            assert_is_one_of(scales, 'scales',
                             ('log', 'log-piecewise', 'linear'))
            if nv is None:
                nv = 32
            if wavelet is None:
                raise ValueError("must set `wavelet` if `scales` isn't array")
            scaletype = scales

        elif isinstance(scales, np.ndarray):
            if scales.squeeze().ndim != 1:
                raise ValueError("`scales`, if array, must be 1D "
                                 "(got shape %s)" % str(scales.shape))
            scaletype, _nv = infer_scaletype(scales)
            if scaletype == 'log':
                if nv is not None and _nv != nv:
                    raise Exception("`nv` used in `scales` differs from "
                                    "`nv` passed (%s != %s)" % (_nv, nv))
                nv = _nv
            elif scaletype == 'log-piecewise':
                nv = _nv  # will be array
            scales = scales.reshape(-1, 1)  # ensure 2D for broadcast ops later

        else:
            raise TypeError("`scales` must be a string or Numpy array "
                            "(got %s)" % type(scales))

        if nv is not None and not isinstance(nv, np.ndarray):
            _assert_positive_integer(nv, 'nv')
            nv = int(nv)
        return scaletype, nv, preset

    scaletype, nv, preset = _process_args(scales, nv, wavelet)
    if isinstance(scales, np.ndarray):
        scales = scales.reshape(-1, 1)
        return (scales if not get_params else
                (scales, scaletype, len(scales), nv))

    #### Compute scales & params #############################################
    min_scale, max_scale = cwt_scalebounds(wavelet, N=N, preset=preset,
                                           use_padded_N=use_padded_N)
    scales = make_scales(N, min_scale, max_scale, nv=nv, scaletype=scaletype,
                         wavelet=wavelet)
    na = len(scales)

    return (scales if not get_params else
            (scales, scaletype, na, nv))


def infer_scaletype(scales):
    """Infer whether `scales` is linearly or exponentially distributed (if latter,
    also infers `nv`). Used internally on `scales` and `ssq_freqs`.

    Returns one of: 'linear', 'log', 'log-piecewise'
    """
    if not isinstance(scales, np.ndarray):
        raise TypeError("`scales` must be a numpy array (got %s)" % type(scales))
    elif scales.dtype not in (np.float32, np.float64):
        raise TypeError("`scales.dtype` must be np.float32 or np.float64 "
                        "(got %s)" % scales.dtype)

    th_log = 1e-15 if scales.dtype == np.float64 else 4e-7
    th_lin = th_log * 1e3  # less accurate for some reason

    if np.mean(np.abs(np.diff(scales, 2, axis=0))) < th_lin:
        scaletype = 'linear'
        nv = None

    elif np.mean(np.abs(np.diff(np.log(scales), 2, axis=0))) < th_log:
        scaletype = 'log'
        # ceil to avoid faulty float-int roundoffs
        nv = int(np.round(1 / np.diff(np.log2(scales), axis=0)[0]))

    elif logscale_transition_idx(scales) is None:
        raise ValueError("could not infer `scaletype` from `scales`; "
                         "`scales` array must be linear or exponential. "
                         "(got diff(scales)=%s..." % np.diff(scales, axis=0)[:4])

    else:
        scaletype = 'log-piecewise'
        nv = nv_from_scales(scales)

    return scaletype, nv


def make_scales(N, min_scale=None, max_scale=None, nv=32, scaletype='log',
                wavelet=None, downsample=3):
    """Recommended to first work out `min_scale` & `max_scale` with
    `cwt_scalebounds`.

    # Arguments:
        N: int
            `len(x)` or `len(x_padded)`.

        min_scale, max_scale: float, float
            Set scale range. Obtained e.g. from `utils.cwt_scalebounds`.

        nv: int
            Number of voices (wavelets) per octave.

        scaletype: str['log', 'log-piecewise', 'linear']
            Scaling kind to make.
            `'log-piecewise'` uses `utils.find_downsampling_scale`.

        wavelet: wavelets.Wavelet
            Used only for `scaletype='log-piecewise'`.

        downsample: int
            Downsampling factor. Used only for `scaletype='log-piecewise'`.

    # Returns:
        scales: np.ndarray
    """
    if scaletype == 'log-piecewise' and wavelet is None:
        raise ValueError("must pass `wavelet` for `scaletype == 'log-piecewise'`")
    if min_scale is None and max_scale is None and wavelet is not None:
        min_scale, max_scale = cwt_scalebounds(wavelet, N)
    else:
        min_scale = min_scale or 1
        max_scale = max_scale or N

    # number of 2^-distributed scales spanning min to max
    na = int(np.ceil(nv * np.log2(max_scale / min_scale)))
    # floor to keep freq-domain peak at or to right of Nyquist
    # min must be more precise, if need integer rounding do on max
    mn_pow = int(np.floor(nv * np.log2(min_scale)))
    mx_pow = mn_pow + na

    if scaletype == 'log':
        scales = 2 ** (np.arange(mn_pow, mx_pow) / nv)

    elif scaletype == 'log-piecewise':
        scales = 2 ** (np.arange(mn_pow, mx_pow) / nv)
        idx = find_downsampling_scale(wavelet, scales)

        if idx is not None:
            # `+downsample - 1` starts `scales2` as continuing from `scales1`
            # at `scales2`'s sampling rate; rest of ops are based on this design,
            # such as `/nv` in ssq, which divides `scales2[0]` by `nv`, but if
            # `scales2[0]` is one sample away from `scales1[-1]`, seems incorrect
            scales1 = scales[:idx]
            scales2 = scales[idx + downsample - 1::downsample]
            scales = np.hstack([scales1, scales2])

    elif scaletype == 'linear':
        # TODO poor scheme (but there may not be any good one)
        min_scale, max_scale = 2**(mn_pow/nv), 2**(mx_pow/nv)
        na = int(np.ceil(max_scale / min_scale))
        scales = np.linspace(min_scale, max_scale, na)

    else:
        raise ValueError("`scaletype` must be 'log' or 'linear'; "
                         "got: %s" % scaletype)
    scales = scales.reshape(-1, 1)  # ensure 2D for broadcast ops later
    return scales


def logscale_transition_idx(scales):
    """Returns `idx` that splits `scales` as `[scales[:idx], scales[idx:]]`.
    """
    scales_diff2 = np.abs(np.diff(np.log(scales), 2, axis=0))
    idx = np.argmax(scales_diff2) + 2
    diff2_max = scales_diff2.max()
    # every other value must be zero, assert it is so
    scales_diff2[idx - 2] = 0

    if not np.any(diff2_max > 100*np.abs(scales_diff2).mean()):
        # everything's zero, i.e. no transition detected
        return None
    elif not np.all(np.abs(scales_diff2) < 1e-14):
        # other nonzero diffs found, more than one transition point
        return None
    else:
        return idx


def nv_from_scales(scales):
    """Infers `nv` from `scales` assuming `2**` scales; returns array
    of length `len(scales)` if `scaletype = 'log-piecewise'`.
    """
    logdiffs = 1 / np.diff(np.log2(scales), axis=0)
    nv = np.vstack([logdiffs[:1], logdiffs])

    idx = logscale_transition_idx(scales)
    if idx is not None:
        nv_transition_idx = np.argmax(np.abs(np.diff(nv, axis=0))) + 1
        assert nv_transition_idx == idx, "%s != %s" % (nv_transition_idx, idx)
    return nv


def find_min_scale(wavelet, cutoff=1):
    """Design the wavelet in frequency domain. `scale` is found to yield
    `scale * xi(scale=1)` such that its last (largest) positive value evaluates
    `wavelet` to `cutoff * max(psih)`. If cutoff > 0, it lands to right of peak,
    else to left (i.e. peak excluded).
    """
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    w_peak, peak = find_maximum(wavelet.fn)
    if cutoff > 0:
        # search to right of peak
        step_start, step_limit = w_peak, 10*w_peak
    else:
        # search to left of peak
        step_start, step_limit = 0, w_peak

    w_cutoff, _ = find_first_occurrence(wavelet.fn, value=abs(cutoff) * peak,
                                        step_start=step_start,
                                        step_limit=step_limit)
    min_scale = w_cutoff / np.pi
    return min_scale


def find_max_scale(wavelet, N, bin_loc=1, bin_amp=1):
    """Finds `scale` such that freq-domain wavelet's amplitude is `bin_amp`
    of maximum at `bin_loc` bin. Set `bin_loc=1` to ensure no lower frequencies
    are lost, but likewise mind redundancy (see `make_scales`).
    """
    wavelet = Wavelet._init_if_not_isinstance(wavelet)

    # get scale at which full freq-domain wavelet is likely to fit
    wc_ct = center_frequency(wavelet, kind='peak-ct', N=N)
    scalec_ct = (4/pi) * wc_ct

    # get freq_domain wavelet, positive half (asm. analytic)
    psih = wavelet(scale=scalec_ct, N=N)[:N//2 + 1]
    # get (radian) frequencies at which it was sampled
    xi = wavelet.xifn(scalec_ct, N)
    # get index of psih's peak
    midx = np.argmax(psih)
    # get index where `psih` attains `bin1_amp` of its max value, to left of peak
    w_bin = xi[np.where(psih[:midx] < psih.max()*bin_amp)[0][-1]]

    # find scale such that wavelet amplitude is `bin_amp` of max at `bin_loc` bin
    max_scale = scalec_ct * (w_bin / xi[bin_loc])
    return max_scale


def find_downsampling_scale(wavelet, scales, span=5, tol=3, method='sum',
                            nonzero_th=.02, nonzero_tol=4., viz=False,
                            viz_last=False):
    """Find `scale` past which freq-domain wavelets are "excessively redundant",
    redundancy determined by `span, tol, method, nonzero_th, nonzero_tol`.

    # Arguments
        wavelet: np.ndarray / wavelets.Wavelet
            CWT wavelet.

        scales: np.ndarray
            CWT scales.

        span: int
            Number of wavelets to cross-correlate at each comparison.

        tol: int
            Tolerance value, works with `method`.

        method: str['any', 'all', 'sum']
            Condition relating `span` and `tol` to determine whether wavelets
            are packed "too densely" at a given cross-correlation, relative
            to "joint peak".

                'any': at least one of wavelet peaks lie `tol` or more bins away
                'all': all wavelet peaks lie `tol` or more bins away
                'sum': sum(distances between wavelet peaks and joint peak) > `tol`

        nonzero_th: float
            Wavelet points as a fraction of respective maxima to consider
            nonzero (i.e. `np.where(psih > psih.max()*nonzero_th)`).

        nonzero_tol: float
            Average number of nonzero points in a `span` group of wavelets above
            which testing is exempted. (e.g. if 5 wavelets have 25 nonzero points,
            average is 5, so if `nonzero_tol=4`, the `scale` is skipped/passed).

        viz: bool (default False)
            Visualize every test for debug purposes.

        viz_last: bool (default True)
            Visualize the failing scale (recommended if trying by hand);
            ignored if `viz=True`.
    """
    def check_group(psihs_peaks, joint_peak, method, tol):
        too_dense = False
        distances = np.abs(psihs_peaks[1] - joint_peak)

        if method == 'any':
            dist_max = distances.max()
            if dist_max < tol:
                too_dense = True
        elif method == 'all':
            dist_satisfied = (distances > tol)
            if not np.all(dist_satisfied):
                too_dense = True
        elif method == 'sum':
            dist_sum = distances.sum()
            if dist_sum < tol:
                too_dense = True
        return too_dense

    def _viz(psihs, psihs_peaks, joint_peak, psihs_nonzeros, i):
        max_nonzero_idx = np.where(psihs_nonzeros)[1].max()

        plot(psihs.T[:max_nonzero_idx + 3], color='tab:blue',
             vlines=(joint_peak, {'color': 'tab:red'}))
        scat(psihs_peaks[1], psihs[psihs_peaks].T, color='tab:red', show=1)

        distances = np.abs(psihs_peaks[1] - joint_peak)
        print("(idx, peak distances from joint peak, joint peak) = "
              "({}, {}, {})".format(i, distances, joint_peak), flush=True)

    assert_is_one_of(method, 'method', ('any', 'all', 'sum'))
    if not isinstance(wavelet, np.ndarray):
        wavelet = Wavelet._init_if_not_isinstance(wavelet)

    Psih = (wavelet if isinstance(wavelet, np.ndarray) else
            wavelet(scale=scales))
    if len(Psih) != len(scales):
        raise ValueError("len(Psih) != len(scales) "
                         "(%s != %s)" % (len(Psih), len(scales)))

    # analytic, drop right half (all zeros)
    Psih = Psih[:, :Psih.shape[1]//2]
    n_scales = len(Psih)
    n_groups = n_scales - span - 1
    psihs_peaks = None

    for i in range(n_groups):
        psihs = Psih[i:i + span]

        psihs_nonzeros = (psihs > nonzero_th*psihs.max(axis=1)[:, None])
        avg_nonzeros = psihs_nonzeros.sum() / span
        if avg_nonzeros > nonzero_tol:
            continue

        psihs_peaks = np.where(psihs == psihs.max(axis=1)[:, None])
        joint_peak = np.argmax(np.prod(psihs, 0))  # mutually cross-correlate

        too_dense = check_group(psihs_peaks, joint_peak, method, tol)
        if too_dense:
            break

        if viz:
            _viz(psihs, psihs_peaks, joint_peak, psihs_nonzeros, i)

    if (viz or viz_last) and psihs_peaks is not None:
        print(("Failing scale: (idx, scale) = ({}, {:.2f})\n"
               "out of max:    (idx, scale) = ({}, {:.2f})"
               ).format(i, float(scales[i]), len(scales) - 1, float(scales[-1])))
        _viz(psihs, psihs_peaks, joint_peak, psihs_nonzeros, i)

    return i if (i < n_groups - 1) else None


def integrate_analytic(int_fn, nowarn=False):
    """Assumes function that's zero for negative inputs (e.g. analytic wavelet),
    decays toward right, and is unimodal: int_fn(t<0)=0, int_fn(t->inf)->0.
    Integrates using trapezoidal rule, from 0 to inf (equivalently).

    Integrates near zero separately in log space (useful for e.g. 1/x).
    """
    def _est_arr(mxlim, N):
        t = np.linspace(mxlim, .1, N, endpoint=False)[::-1]
        arr = int_fn(t)

        max_idx = np.argmax(arr)
        min_neglect_idx = _min_neglect_idx(np.abs(arr[max_idx:]),
                                           th=1e-15) + max_idx
        return arr, t, min_neglect_idx

    def _find_convergent_array():
        mxlims = [1, 20, 80, 160]
        for m, mxlim in zip([1, 1, 4, 8], mxlims):
            arr, t, min_neglect_idx = _est_arr(mxlim, N=10000*m)
            # ensure sufficient decay between peak and right endpoint, and
            # that `arr` isn't a flatline (contains wavelet peak)
            if ((len(t) - min_neglect_idx > 1000 * m) and
                np.sum(np.abs(arr)) > 1e-5):
                break
        else:
            if int_nz < 1e-5:
                raise Exception("Could not find converging or non-negligibly"
                                "-valued bounds of integration for `int_fn`")
            elif not nowarn:
                WARN("Integrated only from 1e-15 to 0.1 in logspace")
        return arr[:min_neglect_idx], t[:min_neglect_idx]

    def _integrate_near_zero():
        # sample `intfn` more finely as it might be extremely narrow near zero.
        # this still doesn't work well as float64 zeros the numerator before /w,
        # but the true integral will be negligibly small most of the time anyway
        # (.001 to .1 may not be negligible, however; better captured by logspace)
        t = np.logspace(-15, -1, 1000)
        arr = int_fn(t)
        return integrate.trapz(arr, t)

    int_nz = _integrate_near_zero()
    arr, t = _find_convergent_array()
    return integrate.trapz(arr, t) + int_nz


def find_max_scale_alt(wavelet, N, min_cutoff=.1, max_cutoff=.8):
    """
    Design the wavelet in frequency domain. `scale` is found to yield
    `scale * xi(scale=1)` such that two of its consecutive values land
    symmetrically about the peak of `psih` (i.e. none *at* peak), while
    still yielding `wavelet(w)` to fall between `min_cutoff`* and `max_cutoff`*
    `max(psih)`. `scale` is selected such that the symmetry is attained
    using smallest possible bins (closest to dc). Steps:

        1. Find `w` (input value to `wavelet`) for which `wavelet` is maximized
        (i.e. peak of `psih`).
        2. Find two `w` such that `wavelet` attains `min_cutoff` and `max_cutoff`
        times its maximum value, using `w` in previous step as upper bound.
        3. Find `div_size` such that `xi` lands at both points of symmetry;
        `div_size` == increment between successive values of
        `xi = scale * xi(scale=1)`.
            - `xi` begins at zero; along the cutoff bounds, and us selecting
            the smallest number of divisions/increments to reach points of
            symmetry, we guarantee a unique `scale`.

    This yields a max `scale` that'll generally lie in 'nicely-behaved' region
    of std_t; value can be used to fine-tune further.
    See `visuals.sweep_std_t`.
    """
    if max_cutoff <= 0 or min_cutoff <= 0:
        raise ValueError("`max_cutoff` and `min_cutoff` must be positive "
                         "(got %s, %s)" % (max_cutoff, min_cutoff))
    elif max_cutoff <= min_cutoff:
        raise ValueError("must have `max_cutoff > min_cutoff` "
                         "(got %s, %s)" % (max_cutoff, min_cutoff))

    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    w_peak, peak = find_maximum(wavelet.fn)

    # we solve the inverse problem; instead of looking for spacing of xi
    # that'd land symmetrically about psih's peak, we pick such points
    # above a set ratio of peak's value and ensure they divide the line
    # from left symmetry point to zero an integer number of times

    # define all points of wavelet from cutoff to peak, left half
    w_cutoff, _ = find_first_occurrence(wavelet.fn, value=min_cutoff * peak,
                                        step_start=0, step_limit=w_peak)

    w_ltp = np.arange(w_cutoff, w_peak, step=1/N)  # left-to-peak

    # consider every point on wavelet(w_ltp) (except peak) as candidate cutoff
    # point, and pick earliest one that yields integer number of increments
    # from left point of symmetry to origin
    div_size = (w_peak - w_ltp[:-1]) * 2  # doubled so peak is skipped
    n_divs = w_ltp[:-1] / div_size
    # diff of modulus; first drop in n_divs is like [.98, .99, 0, .01], so at 0
    # we've hit an integer, and n_divs grows ~linearly so behavior guaranteed
    # -.8 arbitrary to be ~1 but <1
    try:
        idx = np.where(np.diff(n_divs % 1) < -.8)[0][0]
    except:
        raise Exception("Failed to find suffciently-integer xi divisions; try "
                        "widening (min_cutoff, max_cutoff)")
    # the div to base the scale on (angular bin spacing of scale*xi)
    div_scale = div_size[idx + 1]

    # div size of scale=1 (spacing between angular bins at scale=1)
    w_1div = np.pi / (N / 2)

    max_scale = div_scale / w_1div
    return max_scale


def _process_fs_and_t(fs, t, N):
    """Ensures `t` is uniformly-spaced and of same length as `x` (==N)
    and returns `fs` and `dt` based on it, or from defaults if `t` is None.
    """
    if t is not None:
        if len(t) != N:
            # not explicitly used anywhere but ensures wrong `t` wasn't supplied
            raise Exception("`t` must be of same length as `x` "
                            "(%s != %s)" % (len(t), N))
        elif not np.mean(np.abs(np.diff(t, 2, axis=0))) < 1e-7:  # float32 thr.
            raise Exception("Time vector `t` must be uniformly sampled.")
        fs = 1 / (t[1] - t[0])
    else:
        if fs is None:
            fs = 1
        elif fs <= 0:
            raise ValueError("`fs` must be > 0")
    dt = 1 / fs
    return dt, fs, t


#############################################################################
from ..wavelets import Wavelet, center_frequency
from ..visuals import plot, scat, _viz_cwt_scalebounds, wavelet_waveforms
from ..visuals import sweep_harea



# -*- coding: utf-8 -*-
"""Convenience visual methods"""
import numpy as np
import matplotlib.pyplot as plt
from numpy.fft import ifft
from pathlib import Path
from .algos import find_closest, find_maximum
from .configs import gdefaults


#### Visualizations ##########################################################
def wavelet_tf(wavelet, N=2048, scale=None, notext=False, width=1.1, height=1):
    """Visualize `wavelet` joint time-frequency resolution. Plots frequency-domain
    wavelet (psih) along y-axis, and time-domain wavelet (psi) along x-axis.

    Orthogonal units (e.g. y-axis for psi) are meaningless; function values
    aren't to scale, but *widths* are, so time-frequency uncertainties are
    accurately captured.

    `wavelet` is instance of `wavelets.Wavelet` or its valid `wavelet` argument.
    See also: https://www.desmos.com/calculator/nqowgloycy
    """
    def pick_scale(wavelet, N):
        """Pick scale such that both time- & freq-domain wavelets look nice."""
        st_min, st_max = 65 * (N / 2048), 75 * (N / 2048)
        max_iters = 100
        scale = wavelet.scalec_ct
        # generous `min_decay` since we don't care about initial bad cases
        kw = dict(wavelet=wavelet, N=N, min_decay=1, nondim=False)
        std_t = time_resolution(scale=scale, **kw)

        i = 0
        while not (st_min < std_t < st_max):
            if std_t > st_max:
                scale /= 1.1
            else:
                scale *= 1.1
            std_t = time_resolution(scale=scale, **kw)

            if i > max_iters:
                raise ValueError(f"couldn't autofind `scale` after {max_iters} "
                                 "iterations, aborting")
            i += 1
        return scale

    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    if scale is None:
        scale = pick_scale(wavelet, N)

    #### Compute psi & psihf #################################################
    psi  = ifft(wavelet(scale * _xifn(1, N)) * (-1)**np.arange(N))
    apsi = np.abs(psi)
    t = np.arange(-N/2, N/2, step=1)

    w = _xifn(1, N)[:N//2 + 1]
    psih = wavelet(scale * w)

    #### Compute stdevs & respective indices #################################
    wc    = center_frequency(wavelet, scale, N)
    std_w = freq_resolution(wavelet, scale, N, nondim=0)
    std_t = time_resolution(wavelet, scale, N, nondim=0, min_decay=1)
    _wc = np.pi - wc

    wlix = np.argmin(np.abs(w - (_wc - std_w)))
    wrix = np.argmin(np.abs(w - (_wc + std_w)))
    wl, wr = w[wlix], w[wrix]

    tlix = np.argmin(np.abs(t - (0 - std_t)))
    trix = np.argmin(np.abs(t - (0 + std_t)))
    tl, tr = t[tlix], t[trix]

    ## Rescale psi so that its y-coords span 1/5 of psih's x-coords, & vice-versa
    frac = 5
    psig  = psi  * (w.max() / apsi.max()) / frac
    apsig = apsi * (w.max() / apsi.max()) / frac
    psihg = psih * (t.max() / psih.max()) / frac
    # additionally shift psih to psi's left
    psihg += t.min()

    ## Find intersections
    w_xminu, w_xmax = psihg[::-1][wlix], tr
    w_xmind = psihg[::-1][wrix]  # psih not necessarily symmetric
    w_ymin, w_ymax = wl, wr
    t_xmin, t_xmax = tl, tr
    t_yminl, t_ymax = apsig[tlix], wr
    t_yminr = apsig[trix]  # same for psi

    #### Plot ################################################################
    plot(t, psig, complex=1, h=1.5)
    plot(t, apsig, linestyle='--', color='k')
    plot(psihg[::-1], w, color='purple')

    # bounds lines
    lkw = dict(color='k', linewidth=1)
    plot([t_xmin,  t_xmin], [t_yminl, t_ymax], **lkw)
    plot([t_xmax,  t_xmax], [t_yminr, t_ymax], **lkw)
    plot([w_xminu, w_xmax], [w_ymin,  w_ymin], **lkw)
    plot([w_xmind, w_xmax], [w_ymax,  w_ymax], **lkw)
    plt.xlim(t.min()*1.02, t.max()*1.02)

    # radians 0 to pi from top to bottom(=psi's mean)
    ylabels = np.round(np.linspace(np.pi, 0, 7), 1)
    plt.yticks(np.linspace(0, np.pi, len(ylabels)), ylabels)

    if notext:
        plt.gcf().set_size_inches(12*width, 12*height)
        plt.show()
        return
    #### Title, annotations, labels, styling #################################
    ## Annotation: info summary
    txt = ("    wc = {:<6.5f} rad-c/s\n"
           " std_t = {:<6.4f} s/c-rad\n"
           " std_w = {:<6.5f} rad-c/s\n"
           "area/4 = {:.12f}\n"
           "       = std_t * std_w\n\n"
           "(rad-c/s=\n radians*cycles/samples)"
           ).format(wc, std_t, std_w, std_t * std_w)
    _annotate(txt, xy=(.7, .76), fontsize=16)

    ## Title: wavelet name & parameters
    title = wavelet._desc(N=N, scale=scale)
    plt.title(title, loc='left', weight='bold', fontsize=16)

    ## Styling
    plt.xlabel("samples", weight='bold', fontsize=15)
    plt.ylabel("radians", weight='bold', fontsize=15)

    plt.gcf().set_size_inches(12*width, 12*height)
    plt.show()


def wavelet_tf_anim(wavelet, N=2048, scales=None, width=1.1, height=1,
                    savepath='wavanim.gif', testing=False):
    """This method computes same as `wavelet_tf` but for all scales at once,
    and animates 'intelligently'. See help(wavelet_tf).

    `scales=None` will default to 'log:minimal' with (.9*min_scale,
    0.25*max_scale). These are selected to show the wavelet a little outside of
    "well-behaved" range (without slashing max_scale, it's a lot outside such
    range). May not work for every wavelet or all of their configs.
    """
    def _make_anim_scales(scales, wavelet, N):
        if scales is None:
            scales = 'log:minimal'
            mn, mx = cwt_scalebounds(wavelet, N=N, preset='maximal',
                                     use_padded_N=False)
            scales = make_scales(N, 0.90*mn, 0.25*mx, scaletype='log')
        else:
            scales = process_scales(scales, N, wavelet, use_padded_N=False)

        # compute early and late scales more densely as they capture more
        # interesting behavior, so animation will slow down smoothly near ends
        scales = scales.squeeze()
        na = len(scales)

        s0 = (25/253)*na  # empircally-determined good value

        srepl = max(int(s0), 1)  # scales to keep from each end
        srepr = max(int(s0), 1)
        smull = 4        # extension factor
        smulr = 3

        sright = np.linspace(scales[-srepr], scales[-1],    srepr * smulr)
        sleft  = np.linspace(scales[0],      scales[srepl], srepl * smull)
        sright = np.hstack([sright, sright[-1].repeat(smulr*2)])  # smooth loop
        sleft  = np.hstack([sleft[0].repeat(smull*2), sleft])

        scales = np.hstack([sleft, scales[srepl:-srepr], sright])
        scales  = scales.reshape(-1, 1)
        return scales

    from matplotlib.animation import FuncAnimation
    import matplotlib
    matplotlib.use("Agg")
    NOTE("Switched matplotlib to 'Agg' backend for animating")

    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    scales = _make_anim_scales(scales, wavelet, N)

    #### Compute Psi & Psih ##################################################
    Psi = wavelet.psifn(scale=scales, N=N)
    aPsi = np.abs(Psi)
    t = np.arange(-N/2, N/2, step=1)

    w = _xifn(1, N)[:N//2 + 1]
    Psih = wavelet(scales * w)

    #### Compute stdevs & respective indices #################################
    Wc    = np.zeros(len(scales))
    std_W = Wc.copy()
    std_T = Wc.copy()

    for i, scale in enumerate(scales):
        Wc[i]    = center_frequency(wavelet, float(scale), N, kind='energy')
        std_W[i] = freq_resolution( wavelet, float(scale), N, nondim=0)
        std_T[i] = time_resolution( wavelet, float(scale), N, nondim=0,
                                    min_decay=1)
    _Wc = np.pi - Wc

    Wlix = find_closest((_Wc - std_W).reshape(-1, 1), w).squeeze()
    Wrix = find_closest((_Wc + std_W).reshape(-1, 1), w).squeeze()
    Wl, Wr = w[Wlix], w[Wrix]

    Tlix = find_closest(0 - std_T.reshape(-1, 1), t).squeeze()
    Trix = find_closest(0 + std_T.reshape(-1, 1), t).squeeze()
    Tl, Tr = t[Tlix], t[Trix]

    ## Rescale Psi so that its y-coords span 1/5 of Psih's x-coords, & vice-versa
    frac = 5
    Psig  = Psi  * (w.max() / aPsi.max(axis=-1)).reshape(-1, 1) / frac
    aPsig = aPsi * (w.max() / aPsi.max(axis=-1)).reshape(-1, 1) / frac
    Psihg = Psih * (t.max() / Psih.max(axis=-1)).reshape(-1, 1) / frac
    # additionally shift Psih to Psi's left
    Psihg += t.min()

    ## Find intersections ####################################################
    sidx = np.arange(len(scales))

    W_xminu, W_xmax = Psihg[:, ::-1][sidx, Wlix], Tr
    W_xmind = Psihg[:, ::-1][sidx, Wrix]  # Psih not necessarily symmetric
    W_ymin, W_ymax = Wl, Wr

    T_xmin, T_xmax = Tl, Tr
    T_yminl, T_ymax = aPsig[sidx, Tlix], Wr
    T_yminr = aPsig[sidx, Trix]  # same for Psi

    ## Set up plot objects ###################################################
    fig, ax = plt.subplots()
    ax.set_xlim([t.min()*1.02, t.max()*1.02])
    ax.set_ylim([-aPsig.max()*1.05, np.pi*1.02])

    ylabels = np.round(np.linspace(np.pi, 0, 7), 1)
    plt.yticks(np.linspace(0, np.pi, len(ylabels)), ylabels)

    fig.set_size_inches(12*width, 12*height)

    ## Title: wavelet name & parameters
    title = wavelet._desc(N=N)
    ax.set_title(title, loc='left', weight='bold', fontsize=16)

    line1, = ax.plot([], [], color='tab:blue')
    line2, = ax.plot([], [], color='tab:orange')
    line3, = ax.plot([], [], color='k', linestyle='--')
    line4, = ax.plot([], [], color='purple')

    lkw = dict(color='k', linewidth=1)
    line5, = ax.plot([], [], **lkw)
    line6, = ax.plot([], [], **lkw)
    line7, = ax.plot([], [], **lkw)
    line8, = ax.plot([], [], **lkw)

    tkw = dict(horizontalalignment='center', verticalalignment='center',
               transform=ax.transAxes, fontsize=15, weight='bold')
    txt = ax.text(.9, .95, "scale=%.2f" % scales[0], **tkw)
    fig.tight_layout()

    #### Animate #############################################################
    def unique_savepath(savepath):
        """Ensure doesn't overwrite existing"""
        sp = Path(savepath)
        savename = sp.stem

        if sp.is_file():
            paths = [str(p.stem) for p in Path(savepath).parent.iterdir()
                     if savename in p.stem]
            maxnum = 0
            for p in paths:
                num = p.replace(savename, '')
                if num != '' and int(num) > maxnum:
                    maxnum = int(num)
            sp = Path(sp.parent, savename + str(maxnum + 1) + sp.suffix)
        sp = str(sp)
        return sp

    def animate(i):
        line1.set_data(t, Psig[i].real)
        line2.set_data(t, Psig[i].imag)
        line3.set_data(t, aPsig[i])
        line4.set_data(Psihg[i][::-1], w)

        line5.set_data([T_xmin[i],  T_xmin[i]], [T_yminl[i], T_ymax[i]])
        line6.set_data([T_xmax[i],  T_xmax[i]], [T_yminr[i], T_ymax[i]])
        line7.set_data([W_xminu[i], W_xmax[i]], [W_ymin[i],  W_ymin[i]])
        line8.set_data([W_xmind[i], W_xmax[i]], [W_ymax[i],  W_ymax[i]])

        txt.set_text("scale=%.2f" % scales[i])
        return line1, line2, line3, line4, line5, line6, line7, line8

    sp = unique_savepath(savepath)
    print(("Successfully computed parameters, scales ranging {:.2f} to {:.2f}; "
           "animating...\nWill save to: {}").format(
               scales.min(), scales.max(), sp), flush=True)

    frames = np.hstack([range(len(scales)), range(len(scales) - 1)[::-1]])
    if testing:  # animation takes long; skip when unit-testing
        print("Passed `testing=True`, won't animate")
        return
    anim = FuncAnimation(fig, animate, frames=frames, interval=60,
                         blit=True, repeat=False)

    anim.save(sp, writer='imagemagick')
    print("Animated and saved to", sp, flush=True)


def wavelet_heatmap(wavelet, scales='log', N=2048):
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    if not isinstance(scales, np.ndarray):
        scales = process_scales(scales, N, wavelet, use_padded_N=False)

    #### Compute time- & freq-domain wavelets for all scales #################
    Psi = wavelet.psifn(scale=scales, N=N)

    w = _xifn(1, N)[:N//2 + 1]
    Psih = wavelet(scales * w)

    #### Plot ################################################################
    mx = np.abs(Psi).max() * .01
    title0 = wavelet._desc(N=N)

    kw = dict(ylabel="scales", xlabel="samples")
    imshow(Psi.real,   norm=(-mx, mx), yticks=scales,
           title=title0 + " | Time-domain; real part", **kw)

    imshow(Psi, abs=1, cmap='bone', norm=(0, mx), yticks=scales,
           title=title0 + " | Time-domain; abs-val", **kw)

    kw['xlabel'] = "radians"
    imshow(Psih, abs=1, cmap='jet', yticks=scales,
           xticks=np.linspace(0, np.pi, N//2),
           title=title0 + " | Freq-domain; abs-val", **kw)


def sweep_std_t(wavelet, N, scales='log', get=False, **kw):
    def _process_kw(kw):
        kw = kw.copy()  # don't change external dict
        defaults = dict(min_decay=1, max_mult=2, min_mult=2,
                        nondim=False, force_int=True)
        for k, v in kw.items():
            if k not in defaults:
                raise ValueError(f"unsupported kwarg '{k}'; must be one of: "
                                 + ', '.join(defaults))

        for k, v in defaults.items():
            kw[k] = kw.get(k, v)
        return kw

    kw = _process_kw(kw)
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    scales = process_scales(scales, N, wavelet)

    std_ts = np.zeros(scales.size)
    for i, scale in enumerate(scales):
        std_ts[i] = time_resolution(wavelet, scale=scale, N=N, **kw)

    title = "std_t [{}] vs log2(scales) | {} wavelet, {}".format(
        "nondim" if kw['nondim'] else "s/c-rad", wavelet.name, wavelet.config_str)
    hlines = ([N/2, N/4], dict(color='k', linestyle='--'))
    plot(np.log2(scales), std_ts, title=title, hlines=hlines, show=1)

    if get:
        return std_ts


def sweep_std_w(wavelet, N, scales='log', get=False, **kw):
    def _process_kw(kw):
        kw = kw.copy()  # don't change external dict
        defaults = dict(nondim=False, force_int=True)
        for k, v in kw.items():
            if k not in defaults:
                raise ValueError(f"unsupported kwarg '{k}'; must be one of: "
                                 + ', '.join(defaults))

        for k, v in defaults.items():
            kw[k] = kw.get(k, v)
        return kw

    kw = _process_kw(kw)
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    scales = process_scales(scales, N, wavelet)

    std_ws = np.zeros(scales.size)
    for i, scale in enumerate(scales):
        std_ws[i] = freq_resolution(wavelet, scale=scale, N=N, **kw)

    title = "std_w [{}] vs log2(scales) | {} wavelet, {}".format(
        "nondim" if kw['nondim'] else "s/c-rad", wavelet.name, wavelet.config_str)
    plot(np.log2(scales), std_ws, title=title, show=1)

    if get:
        return std_ws


def sweep_harea(wavelet, N, scales='log', get=False, kw_w=None, kw_t=None):
    """Sub-.5 and near-0 areas will occur for very high scales as a result of
    discretization limitations. Zero-areas have one non-zero frequency-domain,
    and std_t==N/2, with latter more accurately set to infinity (which we don't).

    Sub-.5 are per freq-domain assymetries degrading time-domain decay,
    and limited bin discretization integrating unreliably (yet largely
    meaningfully; the unreliable-ness appears emergent from discretization).
    """
    kw_w, kw_t = (kw_w or {}), (kw_t or {})
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    scales = process_scales(scales, N, wavelet)

    std_ws = sweep_std_w(wavelet, N, scales, get=True, **kw_w)
    plt.show()
    std_ts = sweep_std_t(wavelet, N, scales, get=True, **kw_t)
    plt.show()
    hareas = std_ws * std_ts

    hline = (.5, dict(color='tab:red', linestyle='--'))
    title = "(std_w * std_t) vs log2(scales) | {} wavelet, {}".format(
        wavelet.name, wavelet.config_str)
    plot(np.log2(scales), hareas, color='k', hlines=hline, title=title)
    plt.show()

    if get:
        return hareas, std_ws, std_ts


def wavelet_waveforms(wavelet, N, scale, zoom=True):
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    ## Freq-domain sampled #######################
    w_peak, _ = find_maximum(wavelet.fn)

    w_ct = np.linspace(0, w_peak*2, max(4096, p2up(N)[0]))  # 'continuous-time'
    w_dt = np.linspace(0, np.pi, N//2) * scale  # sampling pts at `scale`
    psih_ct = wavelet(w_ct)
    psih_dt = wavelet(w_dt)

    title = ("wavelet(w) sampled by xi at scale={:.2f}, N={} | {} wavelet, {}"
             ).format(scale, N, wavelet.name, wavelet.config_str)
    plot(w_ct, psih_ct, title=title, xlabel="radians")
    scat(w_dt, psih_dt, color='tab:red')

    plt.legend(["psih at scale=1", "sampled at scale=%.2f" % scale], fontsize=13)
    plt.axvline(w_peak, color='tab:red', linestyle='--')
    plt.show()

    ## Freq-domain #######################
    # if peak not near left, don't zoom; same as `if .. (w_peak >= w_dt.max())`
    if not zoom or (np.argmax(psih_dt) > .05 * N/2):
        end = None
    else:
        peak_idx = np.argmax(psih_dt)
        end = np.where(psih_dt[peak_idx:] < 1e-4*psih_dt.max())[0][0]
        end += peak_idx + 3  # +3: give few more indices for visual

    w_dtn = w_dt * (np.pi / w_dt.max())  # norm to span true w
    plot(w_dtn[:end], psih_dt[:end], xlabel="radians",
         title="Freq-domain waveform (psih)" + ", zoomed" * (end is not None))
    scat(w_dtn[:end], psih_dt[:end], color='tab:red', show=1)

    ## Time-domain #######################
    psi = wavelet.psifn(scale=scale, N=N)
    apsi = np.abs(psi)
    t = np.arange(-N/2, N/2, step=1)

    # don't zoom unless there's fast decay
    peak_idx = np.argmax(apsi)
    if not zoom or (apsi.max() / apsi[peak_idx:].min() <= 1e3):
        start, end = 0, None
    else:
        dt = np.where(apsi[peak_idx:] < 1e-3*apsi.max())[0][0]
        start, end = (N//2 - dt, N//2 + dt + 1)

    plot(t[start:end], psi[start:end], complex=1, xlabel="samples",
         title="Time-domain waveform (psi)" + ", zoomed" * (end is not None))
    plot(t[start:end], apsi[start:end], color='k', linestyle='--', show=1)


def _viz_cwt_scalebounds(wavelet, N, min_scale=None, max_scale=None,
                         std_t=None, cutoff=1, stdevs=2, Nt=None):
    """Can be used to visualize time & freq domains separately, where
    `min_scale` refers to scale at which to show the freq-domain wavelet, and
    `max_scale` the time-domain one.
    """
    def _viz_max(wavelet, N, max_scale, std_t, stdevs, Nt):
        if Nt is None:
            Nt = p2up(N)[0]
        if std_t is None:
            # permissive max_mult to not crash visual
            std_t = time_resolution(wavelet, max_scale, N, nondim=False,
                                    min_mult=2, max_mult=2, min_decay=1)

        t = np.arange(-Nt/2, Nt/2, step=1)
        t -= t.mean()
        psi = wavelet.psifn(scale=max_scale, N=len(t))

        plot(t, np.abs(psi)**2, ylims=(0, None),
             title="|Time-domain wavelet|^2, extended (outside dashed)")

        plt.axvline(std_t,          color='tab:red')
        plt.axvline(std_t * stdevs, color='tab:green')
        # mark target (non-extended) frame
        _ = [plt.axvline(v, color='k', linestyle='--') for v in (-N/2, N/2-1)]

        _kw = dict(fontsize=16, xycoords='axes fraction', weight='bold')
        plt.annotate("1 stdev",
                     xy=(.88, .95), color='tab:red',   **_kw)
        plt.annotate("%s stdevs" % stdevs,
                     xy=(.88, .90), color='tab:green', **_kw)
        plt.show()

    def _viz_min(wavelet, N, min_scale, cutoff):
        w = _xifn(1, N)[:N//2 + 1]  # drop negative freqs
        psih = wavelet(min_scale * w, nohalf=True)
        _, mx = find_maximum(wavelet)

        plot(w, psih, title=("Frequency-domain wavelet, positive half "
                             "(cutoff=%s, peak=%.3f)" % (cutoff, mx)))
        plt.axhline(mx * abs(cutoff), color='tab:red')
        plt.show()

    if min_scale is not None:
        _viz_min(wavelet, N, min_scale, cutoff)
    if max_scale is not None:
        _viz_max(wavelet, N, max_scale, std_t, stdevs, Nt)
    if not (min_scale or max_scale):
        raise ValueError("Must set at least one of `min_scale`, `max_scale`")


def wavelet_filterbank(wavelet, N=1024, scales='log', skips=0, title_append=None,
                       positives=False, show=True, get=False):
    """Plot all frequency-domain wavelets, superposed.

    `skips=1` will plot every *other* wavelet, `=2` will skip 2, etc.
    `=0` shows all.

    `title_append`: will `title += title_append` if not None. Must be string.
    Can use to display additional info.

    `positives=True` will show full wavelets as opposed to trimmed at Nyquist.

    `get=True` to return the filter bank (ignores `skip`).
    """
    def _title():
        scaletype = infer_scaletype(scales)[0]
        desc = wavelet._desc(N=N)
        desc = desc.replace(" |", " filterbank |")

        title = "{}, scaletype={}{}".format(desc, scaletype, title_append or '')
        title = _textwrap(title, wrap_len=72)
        return title

    # process `scales` & prepare freq-domain wavelets
    scales = process_scales(scales, N, wavelet)
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    Psih = wavelet(scale=scales, N=N)

    # process `skips`
    Psih_show, scales_show = [], []
    for i, psih in enumerate(Psih):
        if i % (skips + 1) == 0:
            Psih_show.append(psih)
            scales_show.append(scales[i])
    Psih_show = np.vstack(Psih_show).T

    # prepare plot params
    if positives:
        w = None
        xlims = (-N/100, N*1.01)
    else:
        Psih_show = Psih_show[:N//2]
        w = np.linspace(0, np.pi, N//2, endpoint=True)
        xlims = (-np.pi/100, np.pi*1.01)

    # plot
    if positives:
        plt.axvline(N/2, color='tab:red')  # show Nyquist
    plot(w, Psih_show, color='tab:blue', title=_title(), xlims=xlims, show=0,
         xlabel="radians")

    # style
    _, ymax = plt.gca().get_ylim()
    plt.ylim(-ymax/100, ymax*1.03)
    txt = "(min, max)=(%.3f, %.1f)" % (np.min(scales_show), np.max(scales_show))
    _annotate(txt, xy=(.63, .95), fontsize=17)

    if show:
        plt.show()
    if get:
        return Psih


def viz_cwt_higher_order(Wx_k, scales=None, wavelet=None, **imshow_kw):
    if wavelet is not None:
        wavelet = Wavelet._init_if_not_isinstance(wavelet)
        title_append = " | " + wavelet._desc(show_N=False)
    else:
        title_append = ''
    yticks = scales.squeeze() if (scales is not None) else None
    if imshow_kw.get('ticks', 1):
        imshow_kw['yticks'] = imshow_kw.get('yticks', yticks)

    if isinstance(Wx_k, list):
        for k, Wx in enumerate(Wx_k):
            title = "abs(CWT), order={}{}".format(k, title_append)
            imshow(Wx, abs=1, title=title, **imshow_kw)

        Wx_ka = np.mean(np.vstack([Wx_k]), axis=0)
        order_str = ','.join(map(str, range(len(Wx_k))))
        title = "abs(CWT), orders {} avg{}".format(order_str, title_append)
        imshow(Wx_ka, abs=1, title=title, **imshow_kw)

    else:
        title = "abs(CWT), higher-order avg{}".format(title_append)
        imshow(Wx_k, abs=1, title=title, **imshow_kw)


def viz_gmw_orders(N=1024, n_orders=3, scale=5, gamma=3, beta=60,
                   norm='bandpass'):
    wavs = []
    for k in range(n_orders):
        wav = Wavelet(('gmw', dict(gamma=gamma, beta=beta, norm=norm, order=k)))
        wavs.append(wav)

    psihs = [wav(scale=scale)[:N//2 + 1] for wav in wavs]
    psis  = [wav.psifn(scale=scale)      for wav in wavs]
    w = np.linspace(0, np.pi, N//2 + 1, endpoint=True)

    desc = wavs[0]._desc(show_N=False)
    orders_str = ','.join(map(str, range(n_orders)))

    for psih in psihs:
        plot(w, psih, title="Freq-domain, orders=%s | %s" % (orders_str, desc))
    plot([], show=1)

    for k, psi in enumerate(psis):
        plot(psi, complex=1)
        plot(psi, abs=1, color='k', linestyle='--', show=1,
             title=f"Time-domain, order={k} | {desc}")


#### Visual tools ## messy code ##############################################
def imshow(data, title=None, show=1, cmap=None, norm=None, complex=None, abs=0,
           w=None, h=None, ridge=0, ticks=1, aspect='auto', ax=None, fig=None,
           yticks=None, xticks=None, xlabel=None, ylabel=None, **kw):
    if (ax or fig) and complex:
        NOTE("`ax` and `fig` ignored if `complex`")
    ax  = ax  or plt.gca()
    fig = fig or plt.gcf()
    kw['interpolation'] = kw.get('interpolation', 'none')

    if norm is None:
        mx = np.max(np.abs(data))
        vmin, vmax = ((-mx, mx) if not abs else
                      (0, mx))
    else:
        vmin, vmax = norm
    if cmap is None:
        cmap = 'jet' if abs else 'bwr'
    _kw = dict(vmin=vmin, vmax=vmax, cmap=cmap, aspect=aspect, **kw)

    if abs:
        ax.imshow(np.abs(data), **_kw)
    elif complex:
        fig, axes = plt.subplots(1, 2)
        axes[0].imshow(data.real, **_kw)
        axes[1].imshow(data.imag, **_kw)
        plt.subplots_adjust(left=0, right=1, bottom=0, top=1,
                            wspace=0, hspace=0)
    else:
        ax.imshow(data.real, **_kw)

    if w or h:
        fig.set_size_inches(12 * (w or 1), 12 * (h or 1))

    if ridge:
        data_mx = np.where(np.abs(data) == np.abs(data).max(axis=0))
        ax.scatter(data_mx[1], data_mx[0], color='r', s=4)

    if not ticks:
        ax.set_xticks([])
        ax.set_yticks([])
    if xticks is not None or yticks is not None:
        _ticks(xticks, yticks)
    if xlabel is not None:
        ax.set_xlabel(xlabel, weight='bold', fontsize=15)
    if ylabel is not None:
        ax.set_ylabel(ylabel, weight='bold', fontsize=15)

    _maybe_title(title, ax=ax)
    if show:
        plt.show()


def plot(x, y=None, title=None, show=0, ax_equal=False, complex=0, abs=0,
         c_annot=False, w=None, h=None, dx1=False, xlims=None, ylims=None,
         vert=False, vlines=None, hlines=None, xlabel=None, ylabel=None,
         xticks=None, yticks=None, ax=None, fig=None, ticks=True, **kw):
    ax  = ax  or plt.gca()
    fig = fig or plt.gcf()

    if x is None and y is None:
        raise Exception("`x` and `y` cannot both be None")
    elif x is None:
        x = np.arange(len(y))
    elif y is None:
        y = x
        x = np.arange(len(x))

    if vert:
        x, y = y, x
    if complex:
        ax.plot(x, y.real, color='tab:blue', **kw)
        ax.plot(x, y.imag, color='tab:orange', **kw)
        if c_annot:
            _kw = dict(fontsize=15, xycoords='axes fraction', weight='bold')
            ax.annotate("real", xy=(.93, .95), color='tab:blue', **_kw)
            ax.annotate("imag", xy=(.93, .90), color='tab:orange', **_kw)
    else:
        if abs:
            y = np.abs(y)
        ax.plot(x, y, **kw)
    if dx1:
        ax.set_xticks(np.arange(len(x)))

    if vlines:
        vhlines(vlines, kind='v')
    if hlines:
        vhlines(hlines, kind='h')
    if not ticks:
        ax.set_xticks([])
        ax.set_yticks([])
    if xticks is not None or yticks is not None:
        _ticks(xticks, yticks)

    _maybe_title(title, ax=ax)
    _scale_plot(fig, ax, show=show, ax_equal=ax_equal, w=w, h=h,
                xlims=xlims, ylims=ylims, dx1=(len(x) if dx1 else 0),
                xlabel=xlabel, ylabel=ylabel)


def plots(X, Y=None, nrows=None, ncols=None, tight=True, sharex=False,
          sharey=False, skw=None, pkw=None, _scat=0, show=1, **kw):
    """Example:
    X = [[None, np.arange(xc, xc + wl)],
         [None, np.arange(xc + hop, xc + hop + wl)],
         None,
         None]
    Y = [[x, window],
         [x, window],
         xbuf[:, xbc],
         xbuf[:, xbc + 1]]
    pkw = [[{}]*2, [{}]*2, *[{'color': 'tab:green'}]*2]
    plots(X, Y, nrows=2, ncols=2, sharey='row', tight=tight, pkw=pkw)
    """
    def _process_args(X, Y, nrows, ncols, tight, skw, pkw, kw):
        X = X if isinstance(X, list) else [X]
        Y = Y if isinstance(Y, list) else [Y]
        skw = skw or {}
        pkw = pkw or [{}] * len(X)

        if nrows is None and ncols is None:
            nrows, ncols = len(X), 1
        elif nrows is None:
            nrows = max(len(X) // ncols, 1)
        elif ncols is None:
            ncols = max(len(X) // nrows, 1)

        default = dict(left=0, right=1, bottom=0, top=1, hspace=.1, wspace=.05)
        if tight:
            if not isinstance(tight, dict):
                tight = default.copy()
            else:
                for name in default:
                    if name not in tight:
                        tight[name] = default[name]

        kw['w'] = kw.get('w', .8)
        kw['h'] = kw.get('h', .8)  # default 'tight' enlarges plot
        return X, Y, nrows, ncols, tight, skw, pkw, kw

    X, Y, nrows, ncols, tight, skw, pkw, kw = _process_args(
        X, Y, nrows, ncols, tight, skw, pkw, kw)

    _, axes = plt.subplots(nrows, ncols, sharex=sharex, sharey=sharey, **skw)
    for ax, x, y, _pkw in zip(axes.flat, X, Y, pkw):
        if isinstance(x, list):
            for _x, _y, __pkw in zip(x, y, _pkw):
                plot(_x, _y, ax=ax, **__pkw, **kw)
                if _scat:
                    scat(_x, _y, ax=ax, **__pkw, **kw)
        else:
            plot(x, y, ax=ax, **_pkw, **kw)
            if _scat:
                scat(x, y, ax=ax, **_pkw, **kw)

    if tight:
        plt.subplots_adjust(**tight)
    if show:
        plt.show()


def scat(x, y=None, title=None, show=0, ax_equal=False, s=18, w=None, h=None,
         xlims=None, ylims=None, dx1=False, vlines=None, hlines=None, ticks=1,
         complex=False, abs=False, xlabel=None, ylabel=None, ax=None, fig=None,
         **kw):
    ax  = ax  or plt.gca()
    fig = fig or plt.gcf()

    if x is None and y is None:
        raise Exception("`x` and `y` cannot both be None")
    elif x is None:
        x = np.arange(len(y))
    elif y is None:
        y = x
        x = np.arange(len(x))

    if complex:
        ax.scatter(x, y.real, s=s, **kw)
        ax.scatter(x, y.imag, s=s, **kw)
    else:
        if abs:
            y = np.abs(y)
        ax.scatter(x, y, s=s, **kw)
    if not ticks:
        ax.set_xticks([])
        ax.set_yticks([])

    _maybe_title(title, ax=ax)
    if vlines:
        vhlines(vlines, kind='v')
    if hlines:
        vhlines(hlines, kind='h')
    _scale_plot(fig, ax, show=show, ax_equal=ax_equal, w=w, h=h,
                xlims=xlims, ylims=ylims, dx1=(len(x) if dx1 else 0),
                xlabel=xlabel, ylabel=ylabel)


def plotscat(*args, **kw):
    show = kw.pop('show', False)
    plot(*args, **kw)
    scat(*args, **kw)
    if show:
        plt.show()


def hist(x, bins=500, title=None, show=0, stats=0):
    x = np.asarray(x)
    _ = plt.hist(x.ravel(), bins=bins)
    _maybe_title(title)
    if show:
        plt.show()
    if stats:
        mu, std, mn, mx = (x.mean(), x.std(), x.min(), x.max())
        print("(mean, std, min, max) = ({}, {}, {}, {})".format(
            *_fmt(mu, std, mn, mx)))
        return mu, std, mn, mx


def vhlines(lines, kind='v'):
    lfn = plt.axvline if kind=='v' else plt.axhline

    if not isinstance(lines, (list, tuple)):
        lines, lkw = [lines], {}
    elif isinstance(lines, (list, np.ndarray)):
        lkw = {}
    elif isinstance(lines, tuple):
        lines, lkw = lines
        lines = lines if isinstance(lines, (list, np.ndarray)) else [lines]
    else:
        raise ValueError("`lines` must be list or (list, dict) "
                         "(got %s)" % lines)

    for line in lines:
        lfn(line, **lkw)


def _fmt(*nums):
    return [(("%.3e" % n) if (abs(n) > 1e3 or abs(n) < 1e-3) else
             ("%.3f" % n)) for n in nums]

def _ticks(xticks, yticks):
    def fmt(ticks):
        return ("%.d" if all(float(h).is_integer() for h in ticks) else
                "%.2f")

    if yticks is not None:
        idxs = np.linspace(0, len(yticks) - 1, 8).astype('int32')
        yt = [fmt(yticks) % h for h in np.asarray(yticks)[idxs]]
        plt.yticks(idxs, yt)
    if xticks is not None:
        idxs = np.linspace(0, len(xticks) - 1, 8).astype('int32')
        xt = [fmt(xticks) % h for h in np.asarray(xticks)[idxs]]
        plt.xticks(idxs, xt)

def _maybe_title(title, ax=None):
    if title is None:
        return

    title, kw = (title if isinstance(title, tuple) else
                 (title, {}))
    defaults = gdefaults('visuals._maybe_title', get_all=True, as_dict=True)
    for name in defaults:
        kw[name] = kw.get(name, defaults[name])

    if ax:
        ax.set_title(str(title), **kw)
    else:
        plt.title(str(title), **kw)


def _scale_plot(fig, ax, show=False, ax_equal=False, w=None, h=None,
                xlims=None, ylims=None, dx1=False, xlabel=None, ylabel=None):
    xmin, xmax = ax.get_xlim()
    rng = xmax - xmin

    ax.set_xlim(xmin + .018 * rng, xmax - .018 * rng)
    if ax_equal:
        yabsmax = max(np.abs([*ax.get_ylim()]))
        mx = max(yabsmax, max(np.abs([xmin, xmax])))
        ax.set_xlim(-mx, mx)
        ax.set_ylim(-mx, mx)
        fig.set_size_inches(8*(w or 1), 8*(h or 1))
    if xlims:
        ax.set_xlim(*xlims)
    if ylims:
        ax.set_ylim(*ylims)
    if dx1:
        plt.xticks(np.arange(dx1))
    if w or h:
        fig.set_size_inches(14*(w or 1), 8*(h or 1))
    if xlabel is not None:
        plt.xlabel(xlabel, weight='bold', fontsize=15)
    if ylabel is not None:
        plt.ylabel(ylabel, weight='bold', fontsize=15)
    if show:
        plt.show()


def _annotate(txt, xy=(.85, .9), weight='bold', fontsize=16):
    _kw = dict(xycoords='axes fraction', xy=xy, weight=weight, fontsize=fontsize)
    try:
        # 'Consolas' for vertical align
        plt.annotate(txt, family='Consolas', **_kw)
    except:
        plt.annotate(txt, **_kw)  # in case platform lacks 'Consolas'


#############################################################################
from .wavelets import Wavelet, _xifn
from .wavelets import center_frequency, freq_resolution, time_resolution
from .utils.common import NOTE, _textwrap, p2up
from .utils.cwt_utils import process_scales, cwt_scalebounds, make_scales
from .utils.cwt_utils import infer_scaletype



# -*- coding: utf-8 -*-
import numpy as np
from numpy.fft import fft, ifft, ifftshift
from .utils import WARN, adm_cwt, adm_ssq, _process_fs_and_t
from .utils import padsignal, process_scales, logscale_transition_idx
from .algos import replace_at_inf_or_nan
from .wavelets import Wavelet


def cwt(x, wavelet='gmw', scales='log-piecewise', fs=None, t=None, nv=32,
        l1_norm=True, derivative=False, padtype='reflect', rpadded=False,
        vectorized=True, order=0, average=None):
    """Continuous Wavelet Transform, discretized, as described in
    Sec. 4.3.3 of [1] and Sec. IIIA of [2]. Uses a form of discretized
    convolution theorem via wavelets in the Fourier domain and FFT of input.

    # Arguments:
        x: np.ndarray
            Input vector, 1D.

        wavelet: str / tuple[str, dict] / `wavelets.Wavelet`
            Wavelet sampled in Fourier frequency domain.
                - str: name of builtin wavelet. See `ssqueezepy.wavs()`
                  or `Wavelet.SUPPORTED`.
                - tuple: name of builtin wavelet and its configs.
                  E.g. `('morlet', {'mu': 5})`.
                - `wavelets.Wavelet` instance. Can use for custom wavelet.
                  See `help(wavelets.Wavelet)`.

        scales: str['log', 'log-piecewise', 'linear', 'log:maximal', ...]
                / np.ndarray
            CWT scales.
                - 'log': exponentially distributed scales, as pow of 2:
                         `[2^(1/nv), 2^(2/nv), ...]`
                - 'log-piecewise': 'log' except very high `scales` are downsampled
                  to prevent redundancy. This is recommended. See
                  https://github.com/OverLordGoldDragon/ssqueezepy/issues/
                  29#issuecomment-776792726
                - 'linear': linearly distributed scales.
                  !!! this scheme is not recommended; use with caution

            str assumes default `preset` of `'minimal'` for low scales and
            `'maximal'` for high, which can be changed via e.g. 'log:maximal'.
            See `preset` in `help(utils.cwt_scalebounds)`.

        nv: int
            Number of voices (wavelets per octave). Suggested >= 16.

        fs: float / None
            Sampling frequency of `x`. Defaults to 1, which for
            `maprange='maximal'` makes ssq frequencies range from 1/dT to 0.5*fs,
            i.e. as fraction of reference sampling rate up to Nyquist limit;
            dT = total duration (N/fs).
            Used to compute `dt`, which is only used if `derivative=True`.
            Overridden by `t`, if provided.
            Relevant on `t` and `dT`: https://dsp.stackexchange.com/a/71580/50076

        t: np.ndarray / None
            Vector of times at which samples are taken (eg np.linspace(0, 1, n)).
            Must be uniformly-spaced.
            Defaults to `np.linspace(0, len(x)/fs, len(x), endpoint=False)`.
            Used to compute `dt`, which is only used if `derivative=True`.
            Overrides `fs` if not None.

        l1_norm: bool (default True)
            Whether to L1-normalize the CWT, which yields a more representative
            distribution of energies and component amplitudes than L2 (see [3]).
            If False (default True), uses L2 norm.

        derivative: bool (default False)
            Whether to compute and return `dWx`. Requires `fs` or `t`.

        padtype: str / None
            Pad scheme to apply on input. See `help(utils.padsignal)`.
            `None` -> no padding.

        rpadded: bool (default False)
             Whether to return padded Wx and dWx.
             `False` drops the added padding per `padtype` to return Wx and dWx
             of .shape[1] == len(x).

        vectorized: bool (default True)
            Whether to compute quantities for all scales at once, which is
            faster but uses more memory.

        order: int (default 0) / tuple[int] / range
            > 0 computes `cwt` with higher-order GMWs. If tuple, computes
            `cwt` at each specified order. See `help(_cwt.cwt_higher_order)`.

        average: bool / None
            Only used for tuple `order`; see `help(_cwt.cwt_higher_order)`.

    # Returns:
        Wx: [na x n] np.ndarray (na = number of scales; n = len(x))
            CWT of `x`. (rows=scales, cols=timeshifts)
        scales: [na] np.ndarray
            Scales at which CWT was computed.
        dWx: [na x n] np.ndarray
            Returned only if `derivative=True`.
            Time-derivative of the CWT of `x`, computed via frequency-domain
            differentiation (effectively, derivative of trigonometric
            interpolation; see [4]). Implements as described in Sec IIIB of [2].

    # References:
        1. Wavelet Tour of Signal Processing, 3rd ed. S. Mallat.
        https://www.di.ens.fr/~mallat/papiers/WaveletTourChap1-2-3.pdf

        2. The Synchrosqueezing algorithm for time-varying spectral analysis:
        robustness properties and new paleoclimate applications. G. Thakur,
        E. Brevdo, N.-S. Fučkar, and H.-T. Wu.
        https://arxiv.org/abs/1105.0010

        3. Rectification of the Bias in the Wavelet Power Spectrum.
        Y. Liu, X. S. Liang, R. H. Weisberg.
        http://ocg6.marine.usf.edu/~liu/Papers/Liu_etal_2007_JAOT_wavelet.pdf

        4. The Exponential Accuracy of Fourier and Chebyshev Differencing Methods.
        E. Tadmor.
        http://webhome.auburn.edu/~jzl0097/teaching/math_8970/Tadmor_86.pdf

        5. Synchrosqueezing Toolbox, (C) 2014--present. E. Brevdo, G. Thakur.
        https://github.com/ebrevdo/synchrosqueezing/blob/master/synchrosqueezing/
        cwt_fw.m
    """
    def _vectorized(xh, scales, wavelet, pn, derivative):
        Wx = (wavelet(scale=scales, nohalf=False) * pn
              ).astype('complex128')
        if derivative:
            dWx = (1j * wavelet.xi / dt) * Wx

        Wx = ifftshift(ifft(Wx * xh, axis=-1), axes=-1)
        if derivative:
            dWx = ifftshift(ifft(dWx * xh, axis=-1), axes=-1)
        return (Wx, dWx) if derivative else (Wx, None)

    def _for_loop(xh, scales, wavelet, pn, derivative):
        Wx = np.zeros((len(scales), len(xh))).astype('complex128')
        if derivative:
            dWx = Wx.copy()

        for i, scale in enumerate(scales):
            # sample FT of wavelet at scale `a`
            # * pn = freq-domain spectral reversal to center time-domain wavelet
            psih = wavelet(scale=scale, nohalf=False) * pn
            Wx[i] = ifftshift(ifft(psih * xh))

            if derivative:
                dpsih = (1j * wavelet.xi / dt) * psih
                dWx[i] = ifftshift(ifft(dpsih * xh))
        return (Wx, dWx) if derivative else (Wx, None)

    def _process_args(x, scales, nv, fs, t):
        if not isinstance(x, np.ndarray):
            raise TypeError("`x` must be a numpy array (got %s)" % type(x))
        elif x.ndim != 1:
            raise ValueError("`x` must be 1D (got x.ndim == %s)" % x.ndim)

        if np.isnan(x.max()) or np.isinf(x.max()) or np.isinf(x.min()):
            WARN("found NaN or inf values in `x`; will zero")
            replace_at_inf_or_nan(x, replacement=0.)

        if isinstance(scales, np.ndarray):
            nv = None
        N = len(x)
        dt, *_ = _process_fs_and_t(fs, t, N=N)
        return N, nv, dt

    if isinstance(order, (tuple, list, range)) or order > 0:
        kw = dict(wavelet=wavelet, scales=scales, fs=fs, t=t, nv=nv,
                  l1_norm=l1_norm, derivative=derivative, padtype=padtype,
                  rpadded=rpadded, vectorized=vectorized)
        return cwt_higher_order(x, order=order, average=average, **kw)

    N, nv, dt = _process_args(x, scales, nv, fs, t)

    if padtype is not None:
        xp, _, n1, _ = padsignal(x, padtype, get_params=True)
    else:
        xp = x

    # zero-mean `xp`, take to freq-domain
    xp -= xp.mean()
    xh = fft(xp, axis=-1)

    # validate `wavelet`, process `scales`, define `pn` for later
    wavelet = _process_gmw_wavelet(wavelet, l1_norm)
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    scales = process_scales(scales, N, wavelet, nv=nv)
    pn = (-1)**np.arange(len(xp))

    # temporarily adjust `wavlet.N`, take CWT
    wavelet_N_orig = wavelet.N
    wavelet.N = len(xp)
    Wx, dWx = (_vectorized(xh, scales, wavelet, pn, derivative) if vectorized else
               _for_loop(  xh, scales, wavelet, pn, derivative))
    wavelet.N = wavelet_N_orig  # restore

    # handle unpadding, normalization
    if not rpadded and padtype is not None:
        # shorten to pre-padded size
        Wx = Wx[:, n1:n1 + N]
        if derivative:
            dWx = dWx[:, n1:n1 + N]
    if not l1_norm:
        # normalize energy per L2 wavelet norm, else already L1-normalized
        Wx *= np.sqrt(scales)
        if derivative:
            dWx *= np.sqrt(scales)

    return ((Wx, scales, dWx) if derivative else
            (Wx, scales))


def icwt(Wx, wavelet='gmw', scales='log-piecewise', nv=None, one_int=True,
         x_len=None, x_mean=0, padtype='zero', rpadded=False, l1_norm=True):
    """The inverse Continuous Wavelet Transform of `Wx`, via double or
    single integral.

    # Arguments:
        Wx: np.ndarray
            CWT computed via `ssqueezepy.cwt`.

        wavelet: str / tuple[str, dict] / `wavelets.Wavelet`
            Wavelet sampled in Fourier frequency domain.
                - str: name of builtin wavelet. `ssqueezepy.wavs()`
                - tuple[str, dict]: name of builtin wavelet and its configs.
                  E.g. `('morlet', {'mu': 5})`.
                - `wavelets.Wavelet` instance. Can use for custom wavelet.

        scales: str['log', 'linear', 'log:maximal', ...] / np.ndarray
            See help(cwt).

        nv: int / None
            Number of voices. Suggested >= 32. Needed if `scales` isn't array
            (will default to `cwt`'s).

        one_int: bool (default True)
            Whether to use one-integral iCWT or double.
            Current one-integral implementation performs best.
                - True: Eq 2.6, modified, of [3]. Explained in [4].
                - False: Eq 4.67 of [1]. Explained in [5].

        x_len: int / None. Length of `x` used in forward CWT, if different
            from Wx.shape[1] (default if None).

        x_mean: float. mean of original `x` (not picked up in CWT since it's an
            infinite scale component). Default 0.

        padtype: str
            Pad scheme to apply on input, in case of `one_int=False`.
            See `help(utils.padsignal)`.

        rpadded: bool (default False)
            True if Wx is padded (e.g. if used `cwt(, rpadded=True)`).

        l1_norm: bool (default True)
            True if Wx was obtained via `cwt(, l1_norm=True)`.

    # Returns:
        x: np.ndarray
            The signal, as reconstructed from Wx.

    # References:
        1. Wavelet Tour of Signal Processing, 3rd ed. S. Mallat.
        https://www.di.ens.fr/~mallat/papiers/WaveletTourChap1-2-3.pdf

        2. The Synchrosqueezing algorithm for time-varying spectral analysis:
        robustness properties and new paleoclimate applications. G. Thakur,
        E. Brevdo, N.-S. Fučkar, and H.-T. Wu.
        https://arxiv.org/abs/1105.0010

        3. Synchrosqueezed Wavelet Transforms: a Tool for Empirical Mode
        Decomposition. I. Daubechies, J. Lu, H.T. Wu.
        https://arxiv.org/pdf/0912.2437.pdf

        4. One integral inverse CWT. OverLordGoldDragon.
        https://dsp.stackexchange.com/a/71274/50076

        5. Inverse CWT derivation. OverLordGoldDragon.
        https://dsp.stackexchange.com/a/71148/50076

        6. Synchrosqueezing Toolbox, (C) 2014--present. E. Brevdo, G. Thakur.
        https://github.com/ebrevdo/synchrosqueezing/blob/master/synchrosqueezing/
        synsq_cwt_fw.m
    """
    #### Prepare for inversion ###############################################
    na, n = Wx.shape
    x_len = x_len or n
    if not isinstance(scales, np.ndarray) and nv is None:
        nv = 32  # must match forward's; default to `cwt`'s

    wavelet = _process_gmw_wavelet(wavelet, l1_norm)
    wavelet = Wavelet._init_if_not_isinstance(wavelet)
    # will override `nv` to match `scales`'s
    scales, scaletype, _, nv = process_scales(scales, x_len, wavelet, nv=nv,
                                              get_params=True)
    assert (len(scales) == na), "%s != %s" % (len(scales), na)

    #### Handle piecewise scales case ########################################
    # `nv` must be left unspecified so it's inferred automatically from `scales`
    # in `process_scales` for each piecewise case
    if scaletype == 'log-piecewise':
        kw = dict(wavelet=wavelet, one_int=one_int, x_len=x_len, x_mean=x_mean,
                  padtype=padtype, rpadded=rpadded, l1_norm=l1_norm)

        idx = logscale_transition_idx(scales)
        x  = icwt(Wx[:idx], scales=scales[:idx], **kw)
        x += icwt(Wx[idx:], scales=scales[idx:], **kw)
        return x
    ##########################################################################

    #### Invert ##############################################################
    if one_int:
        x = _icwt_1int(Wx, scales, scaletype, l1_norm)
    else:
        x = _icwt_2int(Wx, scales, scaletype, l1_norm,
                       wavelet, x_len, padtype, rpadded)

    # admissibility coefficient
    Cpsi = (adm_ssq(wavelet) if one_int else
            adm_cwt(wavelet))
    if scaletype == 'log':
        # Eq 4.67 in [1]; Theorem 4.5 in [1]; below Eq 14 in [2]
        # ln(2**(1/nv)) == ln(2)/nv == diff(ln(scales))[0]
        x *= (2 / Cpsi) * np.log(2 ** (1 / nv))
    else:
        x *= (2 / Cpsi)

    x += x_mean  # CWT doesn't capture mean (infinite scale)
    return x


def _icwt_2int(Wx, scales, scaletype, l1_norm, wavelet, x_len,
               padtype='zero', rpadded=False):
    """Double-integral iCWT; works with any(?) wavelet."""
    # add CWT padding if it doesn't exist
    if not rpadded:
        Wx, n_up, n1, _ = padsignal(Wx, padtype=padtype, get_params=True)

    # see help(cwt) on `norm` and `pn`
    norm = _icwt_norm(scaletype, l1_norm)
    pn = (-1)**np.arange(n_up)
    x = np.zeros(n_up)

    for scale, Wx_scale in zip(scales, Wx):  # TODO vectorize?
        psih = wavelet(scale=scale, N=n_up) * pn
        xa = ifftshift(ifft(fft(Wx_scale) * psih))  # convolution theorem
        x += xa.real / norm(scale)

    x = x[n1:n1 + x_len]  # keep the unpadded part
    return x


def _icwt_1int(Wx, scales, scaletype, l1_norm):
    """One-integral iCWT; assumes analytic wavelet."""
    norm = _icwt_norm(scaletype, l1_norm)
    return (Wx.real / norm(scales)).sum(axis=0)


def _icwt_norm(scaletype, l1_norm):
    if l1_norm:
        norm = ((lambda scale: 1) if scaletype == 'log' else
                (lambda scale: scale))
    else:
        if scaletype == 'log':
            norm = lambda scale: scale**.5
        elif scaletype == 'linear':
            norm = lambda scale: scale**1.5
    return norm


def _process_gmw_wavelet(wavelet, l1_norm):
    """Ensure `norm` for GMW is consistent with `l1_norm`."""
    norm = 'bandpass' if l1_norm else 'energy'

    if isinstance(wavelet, str) and wavelet.lower()[:3] == 'gmw':
        wavelet = ('gmw', {'norm': norm})

    elif isinstance(wavelet, tuple) and wavelet[0].lower()[:3] == 'gmw':
        wavelet, wavopts = wavelet
        wavopts['norm'] = wavopts.get('norm', norm)
        wavelet = (wavelet, wavopts)

    elif isinstance(wavelet, Wavelet):
        if wavelet.name == 'GMW L2' and l1_norm:
            raise ValueError("using GMW L2 wavelet with `l1_norm=True`")
        elif wavelet.name == 'GMW L1' and not l1_norm:
            raise ValueError("using GMW L1 wavelet with `l1_norm=False`")
    return wavelet


def cwt_higher_order(x, wavelet='gmw', order=1, average=None, **kw):
    """Compute `cwt` with GMW wavelets of order 0 to `order`. See `help(cwt)`.

    Yields lower variance and more noise robust representation. See VI in ref[1].

    # Arguments:
        x: np.ndarray
            Input, 1D.

        wavelet: str / wavelets.Wavelet
            CWT wavelet.

        order: int / tuple[int] / range
            Order of GMW to use for CWT. If tuple, will compute for each
            order specified in tuple, subject to `average`.

        average: bool (default True if `order` is tuple)
            If True, will take arithmetic mean of resulting `Wx` (and `dWx`
            if `derivative=True`), else return as list. Note for phase transform,
            one should compute derivative of averaged `Wx` rather than take
            average of individual `dWx`s.
            Ignored with non-tuple `order.

        kw: dict / kwargs
            Arguments to `cwt`.
            If `scales` is string, will reuse zeroth-order's; zeroth order
            isn't included in `order`, will set from wavelet at `order=0`.

    # References
        [1] Generalized Morse Wavelets. S. C. Olhede, A. T. Walden. 2002.
        https://spiral.imperial.ac.uk/bitstream/10044/1/1150/1/
        OlhedeWaldenGenMorse.pdf
    """
    def _process_wavelet(wavelet, order):
        wavelet = Wavelet._init_if_not_isinstance(wavelet)
        if not wavelet.name.lower().startswith('gmw'):
            raise ValueError("`wavelet` must be GMW for higher-order transforms "
                             "(got %s)" % wavelet.name)

        wavopts = wavelet.config.copy()
        wavopts.pop('order')
        wavelets = [Wavelet(('gmw', dict(order=k, **wavopts))) for k in order]
        return wavelets, wavopts

    def _process_args(wavelet, order, average, kw):
        if isinstance(order, (list, range)):
            order = tuple(order)
        if not isinstance(order, (list, tuple)):
            order = [order]
        if len(order) == 1 and average:
            WARN("`average` ignored with single `order`")
            average = False

        wavelets, wavopts = _process_wavelet(wavelet, order)

        scales = kw.get('scales', 'log-piecewise')
        if isinstance(scales, str):
            wav = Wavelet(('gmw', dict(order=0, **wavopts)))
            scales = process_scales(scales, len(x), wavelet=wav,
                                    nv=kw.get('nv', 32))
        kw['scales'] = scales

        return wavelets, order, average

    wavelets, order, average = _process_args(wavelet, order, average, kw)

    Wx_all = []
    derivative = kw.get('derivative', False)
    if derivative:
        dWx_all = []

    # take the CWTs
    for k in range(len(order)):
        out = cwt(x, wavelets[k], order=0, **kw)
        Wx_all.append(out[0])
        if derivative:
            dWx_all.append(out[-1])

    # handle averaging; strip `Wx_all` of list container if only one array
    if average or (average is None and isinstance(order, tuple)):
        Wx_all = np.mean(np.vstack([Wx_all]), axis=0)
        if derivative:
            dWx_all = np.mean(np.vstack([dWx_all]), axis=0)
    elif len(Wx_all) == 1:
        Wx_all = Wx_all[0]
        if derivative:
            dWx_all = dWx_all[0]

    return ((Wx_all, kw['scales'], dWx_all) if derivative else
            (Wx_all, kw['scales']))



# -*- coding: utf-8 -*-
"""
Signals for testing effectiveness of time-frequency transforms against
variety of localization characteristics.

1. **sine**: pure sine or cosine at one frequency, `cos(2pi f t)`
    a. sine
    b. cosine
    c. phase-shifted
    d. trimmed (others complete exactly one cycle) (not implemented but is
       trivial; do e.g. `x = x[20:-50]`)

2. **<name>:am**: <name> with amplitude modulation, i.e. `A(t) * fn(t)`
    a. |sine|
    b. |cosine|
    c. exp
    d. gauss

3. **#<name>**: superimpose reflected <name> onto itself, i.e. `x += x[::-1]`

4. **lchirp**: linear chirp, `cos(2pi t**2/2)`, spanning `fmin` to `fmax`

5. **echirp**: exponential chirp, `cos(2pi exp(t))`, spanning `fmin` to `fmax`

6. **hchirp**: hyperbolic chirp, `cos(2pi a/(b - t))`, spanning `fmin` to `fmax`

7, 8, 9: **par_lchirp, par_echirp, par_hchirp**: linear, exponential, hyperbolic
         chirps, superposed, with frequency modulation in parallel,
         spanning `fmin1` to `fmax1` and `fmin2` to `fmax2`.

10. **jumps**: large instant frequency transitions, `cos(2pi f*t), f=2 -> f=100`

11. **packed**: closely-spaced bands of sinusoids with majority overlap, e.g.
                `cos(w*t[No:]) + cos((w+1)*t[-No:]) + cos((w+3)*t[No:]) + ...`,
                `No = .8*len(t)`.

12. **packed_poly**: closely-packed polynomial frequency modulations
                (non-configurable)
                Generates https://www.desmos.com/calculator/swbhgezpjk with A.M.

13. **poly_cubic**: cubic polynomial frequency variation + pure tone
               (non-configurable)
"""
import inspect
import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as sig
from numpy.fft import rfft

from ._ssq_cwt import ssq_cwt
from ._ssq_stft import ssq_stft
from .utils import WARN, _textwrap
from .wavelets import Wavelet
from .visuals import plot, plots, imshow
from .ridge_extraction import extract_ridges


pi = np.pi
DEFAULT_N = 512
DEFAULT_SNR = None
DEFAULT_SEED = None
DEFAULT_ARGS = {
    'cosine': dict(f=64, phi0=0),
    'sine':   dict(f=64, phi0=0),
    'lchirp': dict(tmin=0, tmax=1, fmin=0, fmax=None),
    'echirp': dict(tmin=0, tmax=1, fmin=1, fmax=None),
    'hchirp': dict(tmin=0, tmax=1, fmin=1, fmax=None),
    'jumps':  dict(),
    'low':    dict(),
    'am-cosine': dict(amin=.1),
    'am-sine':   dict(amin=.1),
    'am-exp':    dict(amin=.1),
    'am-gauss':  dict(amin=.01),
    'sine:am-cosine': (dict(f=16), dict(amin=.5)),
}
DEFAULT_TKW = dict(tmin=0, tmax=1, endpoint=True)


#### Test signals ############################################################
class TestSignals():
    """Signals of varying time-frequency characteristics. Convenience methods
    to plot multiple signals and their transforms under varying wavelet / window
    parameters.

    `.demo(signals)` to visualize `signals`, `test_transforms(fn)` to apply `fn`
    to and visualize output.

    See `examples/` on Github, and
    https://overlordgolddragon.github.io/test-signals/

    Also see `help(ssqueezepy._test_signals)`, `TestSignals.SUPPORTED`,
    `TestSignals.DEMO`.

    **Sweep functions**
        For `lchirp`, `echirp`, & `hchirp`, `N` will be determined automatically
        if `tmin`, `tmax`, `fmin`, and `fmax` are provided, minimally such that
        no aliasing occurs.

    **Demo signals**
        `TestSignals.DEMO` holds list of `signals` names invoked when passing
        `signals='all'`, which can be changed.

    # Arguments
        N: int
            Will use this as default `N` anytime `N` is left unspecified.

        snr: float / None
            If not None, will add random normal (White Gaussian) noise to
            signal of SNR `snr` - computed as:
                SNR = 10*log10(xo_var / noise_var)
                noise_var = xo_var / 10^(SNR/10)
            where `xo_var` = unnoised signal variance.

        default_args: dict
            `{<signal_name>: {'param_name': value}}` pairs, where `signal_name`
            is one of `SUPPORTED`. See `test_signals.DEFAULT_ARGS`.

        default_tkw: dict
            Example with all key-value pairs: `dict(tmin=0, tmax=1)`.

        warn_alias: bool (default True)
            Whether to print warning if generated signal aliases (f > fs/2);
            to disable, pass `warn_alias=False` to `__init__()`, or set directly
            on instance (`TestSignals().warn_alias=False`).

        seed: int / None
            If not None, will `np.random.seed(seed)` before applying `snr` noise.
    """
    SUPPORTED = ['cosine', 'sine', 'lchirp', 'echirp', 'echirp_pc', 'hchirp',
                 'par-lchirp', 'par-echirp', 'par-hchirp', 'jumps', 'packed',
                 'packed-poly', 'poly-cubic',
                 'am-sine', 'am-cosine', 'am-exp', 'am-gauss']
    # what to show with `signal='all'`, and in what order
    DEMO = ['cosine', 'sine',
            'lchirp', 'echirp', 'hchirp',
            '#lchirp', '#echirp', '#hchirp',
            'par-lchirp', 'par-echirp', 'par-hchirp', '#par-lchirp',
            'jumps', 'packed', 'packed-poly', 'poly-cubic',
            'am-sine', 'am-cosine', 'am-exp', 'am-gauss']

    def __init__(self, N=None, snr=None, default_args=None, default_tkw=None,
                 warn_alias=True, seed=None):
        self.N = N    or DEFAULT_N
        self.snr = snr or DEFAULT_SNR
        self.default_args = default_args or DEFAULT_ARGS
        self.default_tkw  = default_tkw  or DEFAULT_TKW
        self.warn_alias   = warn_alias
        self.seed = seed or DEFAULT_SEED

        # set defaults on unspecified
        for k, v in DEFAULT_ARGS.items():
            self.default_args[k] = self.default_args.get(k, v)
        for k, v in DEFAULT_TKW.items():
            self.default_tkw[k] = self.default_tkw.get(k, v)

    #### test signals ########################################################
    def _maybe_warn_alias(self, phi, tol=.02):
        # allow non-trivial overshoot as it may occur but not worth warning
        if self.warn_alias:
            fmax = np.diff(phi).max()
            if (fmax - np.pi) > tol:
                WARN("`%s` has aliased w/ max(diff(phi))=%.6f>%.6f" % (
                    inspect.stack()[2][3], fmax, pi))

    def sine(self, N=None, f=1, phi0=0, **tkw):
        """sin(2pi*f*t + phi)"""
        tkw['endpoint'] = tkw.get('endpoint', False)
        t, *_ = self._process_params(N, tkw)

        phi = 2*pi * f * t + phi0
        self._maybe_warn_alias(phi)
        return np.sin(phi), t

    def cosine(self, N=None, f=1, phi0=0, **tkw):
        """cos(2pi*f*t + phi)"""
        tkw['endpoint'] = tkw.get('endpoint', False)
        t, *_ = self._process_params(N, tkw)

        phi = 2*pi * f * t + phi0
        self._maybe_warn_alias(phi)
        return np.cos(phi), t

    def _generate(self, fn, N, fmin, fmax, **tkw):
        """Used by chirps."""
        t, tmin, tmax, fmax = self._process_params(N, tkw, fn, fmin, fmax)
        phi = fn(t, tmin, tmax, fmin, fmax)
        self._maybe_warn_alias(phi)
        return np.cos(phi), t

    def lchirp(self, N=None, fmin=0, fmax=None, **tkw):
        """
        >>>   f(t) = a*t + b
        >>> phi(t) = (a/2)*(t^2 - tmin^2) + b*(t - tmin)
        >>> a = (fmin - fmax) / (tmin - tmax)
            b = (fmin*tmax - fmax*tmin) / (tmax - tmin)
        """
        return self._generate(self._lchirp_fn, N, fmin, fmax, **tkw)

    def _lchirp_fn(self, t, tmin, tmax, fmin, fmax, get_w=False):
        a = (fmin - fmax) / (tmin - tmax)
        b = (fmin*tmax - fmax*tmin) / (tmax - tmin)

        phi = (a/2)*(t**2 - tmin**2) + b*(t - tmin)
        phi *= (2*pi)
        if get_w:
            w = a*t + b
            w *= (2*pi)
        return (phi, w) if get_w else phi

    def echirp(self, N=None, fmin=.1, fmax=None, **tkw):
        """
        >>> f(t)   = a*b^t
        >>> phi(t) = (a/ln(b)) * (b^t - b^tmin)
        >>> a = (fmin^tmax / fmax^tmin) ^ 1/(tmax - tmin)
            b = fmax^(1/tmax) * (1/a)^(1/tmax)
        """
        return self._generate(self._echirp_fn, N, fmin, fmax, **tkw)

    def _echirp_fn(self, t, tmin, tmax, fmin, fmax, get_w=False):
        a = (fmin**tmax / fmax**tmin) ** (1/(tmax - tmin))
        b = fmax**(1/tmax) * (1/a)**(1/tmax)

        phi = (a/np.log(b)) * (b**t - b**tmin)
        phi *= (2*pi)
        if get_w:
            w = a*b**t
            w *= (2*pi)
        return (phi, w) if get_w else phi

    def echirp_pc(self, N=None, fmin=0, fmax=None, **tkw):
        """Alternate design that keeps f'(t) fixed at `e`, but is no longer
        geometric in the sense `f(t2) / f(t1) = const.`. "echirp plus constant".

        >>> f(t)   = a*exp(t) + b
        >>> phi(t) = a*(exp(t) - exp(tmin)) + b*(t - tmin)
        >>> a = (fmax - fmin)/(exp(tmax) - exp(tmin))
            b = (fmin*exp(tmax) - fmax*exp(tmin)) / (exp(tmax) - exp(tmin))
        """
        return self._generate(self._echirp_pc_fn, N, fmin, fmax, **tkw)

    def _echirp_pc_fn(self, t, tmin, tmax, fmin, fmax, get_w=False):
        a, b, c, d = fmin, fmax, tmin, tmax
        A = (b - a) / (np.exp(d) - np.exp(c))
        B = (a*np.exp(d) - b*np.exp(c)) / (np.exp(d) - np.exp(c))

        phi = A*(np.exp(t) - np.exp(tmin)) + B*(t - tmin)
        phi *= (2*pi)
        if get_w:
            w = A*np.exp(t) + B
            w *= (2*pi)
        return (phi, w) if get_w else phi

    def hchirp(self, N=None, fmin=.1, fmax=None, **tkw):
        """
        >>> f(t)   = A / (B - t)^2
        >>> phi(t) = A * (1/(B - t) + 1/(tmin - B))
        >>> a, b, c, d = fmin, fmax, tmin, tmax
            A = AN / AD, B = BN / BD,
            AN = 2*sqrt(a^3*b^3*(c - d)^4) + a^2*b*(c - d)^2 + a*b^2*(c - d)^2
            AD = (a - b)^2
            BN = sqrt(a^3*b^3*(c-d)^4) + a^2*b*c*(c-d) + a*b^2*d*(d - c)
            BD = a*b*(a - b)*(c - d)
        """
        return self._generate(self._hchirp_fn, N, fmin, fmax, **tkw)

    def _hchirp_fn(self, t, tmin, tmax, fmin, fmax, get_w=False):
        a, b, c, d = fmin, fmax, tmin, tmax

        AN = (2*np.sqrt(a**3*b**3*(c - d)**4) + a**2*b*(c - d)**2
              + a*b**2*(c - d)**2)
        AD = (a - b)**2
        BN = np.sqrt(a**3*b**3*(c - d)**4) + a**2*b*c*(c - d) + a*b**2*d*(d - c)
        BD = a*b*(a - b)*(c - d)
        A = AN / AD
        B = BN / BD

        phi = A * (1/(B - t) + 1/(tmin - B))
        phi *= (2*pi)
        if get_w:
            w = A / (B - t)**2
            w *= (2*pi)
        return (phi, w) if get_w else phi

    def par_lchirp(self, N=None, fmin1=None, fmax1=None, fmin2=None, fmax2=None,
                   **tkw):
        """Linear frequency modulation in parallel. Should have
        `fmax2 > fmax1`, `fmin2 > fmin1`, and shared `tmin`, `tmax`.
        """
        N = N or self.N
        fdiff_default = N/10

        if fmin1 is None:
            fmin1 = self.default_args['lchirp'].get('fmin', 0)
        if fmin2 is None:
            fmin2 = fmin1 + fdiff_default
        if fmax2 is None or fmax1 is None:
            if fmax1 is None:
                fmax2 = N/2
                fmax1 = fmax2 - fdiff_default
            else:
                fmax2 = min(N/2, fmax1 + fdiff_default)

        x1, t = self.lchirp(N, fmin1, fmax1, **tkw)
        x2, _ = self.lchirp(N, fmin2, fmax2, **tkw)
        x = x1 + x2
        return x, t

    def par_echirp(self, N=None, fmin1=None, fmax1=None, fmin2=None, fmax2=None,
                   **tkw):
        """Exponential frequency modulation in parallel. Should have
        `fmax2 > fmax1`, `fmin2 > fmin1`, and shared `tmin`, `tmax`.
        """
        N = N or self.N
        fratio_default = 1.5

        if fmin1 is None:
            fmin1 = self.default_args['echirp'].get('fmin', 1)
        if fmin2 is None:
            fmin2 = fmin1 * fratio_default
        if fmax2 is None or fmax1 is None:
            if fmax1 is None:
                fmax2 = N/2
                fmax1 = fmax2 / fratio_default
            else:
                fmax2 = min(N/2, fmax1 * fratio_default)

        x1, t = self.echirp(N, fmin1, fmax1, **tkw)
        x2, _ = self.echirp(N, fmin2, fmax2, **tkw)
        x = x1 + x2
        return x, t

    def par_hchirp(self, N=None, fmin1=None, fmax1=None, fmin2=None, fmax2=None,
                   **tkw):
        """Hyperbolic frequency modulation in parallel. Should have
        `fmax2 > fmax1`, `fmin2 > fmin1`, and shared `tmin`, `tmax`.
        """
        N = N or self.N
        fratio_default = 3

        if fmin1 is None:
            fmin1 = self.default_args['hchirp'].get('fmin', 1)
        if fmin2 is None:
            fmin2 = fmin1 * fratio_default
        if fmax2 is None or fmax1 is None:
            if fmax1 is None:
                fmax2 = N/2
                fmax1 = fmax2 / fratio_default
            else:
                fmax2 = min(N/2, fmax1 * fratio_default)

        x1, t = self.hchirp(N, fmin1, fmax1, **tkw)
        x2, _ = self.hchirp(N, fmin2, fmax2, **tkw)
        x = x1 + x2
        return x, t

    def am_sine(self, N=None, f=1, amin=0, amax=1, phi=0, **tkw):
        """Sine amplitude modulation, `|sin(w) + 1| / 2`."""
        N = N or self.N
        _A, t = self.sine(N, f, phi, **tkw)
        _A = (_A + 1) / 2
        return amin + (amax - amin) * _A, t

    def am_cosine(self, N=None, f=1, amin=0, amax=1, phi=0, **tkw):
        """Cosine amplitude modulation, `|cos(w) + 1| / 2`."""
        N = N or self.N
        _A, t = self.cosine(N, f, phi, **tkw)
        _A = (_A + 1) / 2
        return amin + (amax - amin) * _A, t

    def am_exp(self, N=None, amin=.1, amax=1, **tkw):
        """Uses `echirp`'s expression for `f(t)`."""
        N = N or self.N
        t, tmin, tmax = self._process_params(N, tkw)
        _A = self._echirp_fn(t, tmin, tmax, amin, amax, get_w=True)[1]
        _A /= (2*pi)
        return _A, t

    def am_gauss(self, N=None, amin=.1, amax=1, **tkw):
        """Gaussian centered at center sample (`N/2`)."""
        N = N or self.N
        t = _t(-1, 1, N)
        _A = np.exp( -((t - t.mean())**2 * 5) )
        return amin + (amax - amin)*_A, t

    def jumps(self, N=None, freqs=None, **tkw):
        """Large instant freq transitions, e.g. `cos(2pi f*t), f=2 -> f=100`."""
        N = N or self.N
        t, tmin, tmax = self._process_params(N, tkw)

        n_freqs = len(freqs) if freqs is not None else 4
        M = N // n_freqs
        if freqs is None:
            freqs = [1, M/4, M/2, M/16]

        tdiff = tmax - tmin
        x_freqs = []
        endpoint = tkw.get('endpoint', self.default_tkw.get('endpoint', False))

        t_all = _t(tmin, tdiff * len(freqs), M * len(freqs), endpoint)
        for i, f in enumerate(freqs):
            t = t_all[i*M : (i+1)*M]
            x_freqs.append(np.cos(2*pi * f * t))
        x, t = np.hstack(x_freqs), t_all

        return x, t

    def packed(self, N=None, freqs=None, overlap=.8, **tkw):
        """Closely-spaced bands of sinusoids with majority overlap, e.g.
            `cos(w*t[No:]) + cos((w+1)*t[-No:]) + cos((w+3)*t[No:]) + ...`,
            `No = .8*len(t)`.
        """
        N = N or self.N
        t, *_ = self._process_params(N, tkw)
        if freqs is None:
            freqs = [.5, 1, 2, N/10, N/10 + N/50, N/10 + N/25,
                     N/5, N/4, N/3, N/3 + N/10]
        N_overlap = int(overlap*len(t))

        x = np.zeros(len(t))
        for i, f in enumerate(freqs):
            idxs = (slice(0, N_overlap) if (i % 2 == 0) else
                    slice(-N_overlap, None))
            x[idxs] += np.cos(2*pi * f * t[idxs])
        return x, t

    def packed_poly(self, N=None, **tkw):
        """Closely-packed polynomial frequency modulations (non-configurable;
        adjusts with N to keep bands approx unmoved in time-frequency plane).

        Generates https://www.desmos.com/calculator/swbhgezpjk with A.M.
        """
        N = N or self.N
        t = np.linspace(0, 10, N)

        k1, k2, k3 = 10, 2.4, 4.8  # offsets
        adj = N / 512  # keep FMs around same part of time-freq plane
        k1, k2, k3 = k1*adj, k2*adj, k3*adj

        x1 = (1 + .3 * np.cos(t)
              ) * np.cos(2*np.pi * (k1*t - 0.3*adj*np.sin(t) - 1.8*adj*t**1.5))
        x2 = (1 + .2 * np.cos(2*t)) * np.exp(-t/15) * np.cos(
            2*np.pi * (k2*t + 0.5*adj*t**1.2 + .3*np.sin(t)))
        x3 = np.cos(2*np.pi * (k3*t + .2*adj*t**1.3))

        x = x1 + x2 + x3
        return x, t

    def poly_cubic(self, N=None, **tkw):
        """Cubic polynomial frequency variation + pure tone (non-configurable;
        adjusts with N to keep bands approx unmoved in time-frequency plane).
        """
        N = N or self.N
        t  = np.linspace(0, 10, N, endpoint=True)

        p1 = np.poly1d([0.025, -0.36, 1.25, 2.0]) * (N / 256)
        p3 = np.poly1d([0.01, -0.25, 1.5, 4.0]) * (N / 256)
        x1 = sig.sweep_poly(t, p1)
        x3 = sig.sweep_poly(t, p3)
        x2 = np.sin(2*np.pi * (.5*N/256) * t)

        x = x1 + x2 + x3
        return x, t

    #### Test functions ######################################################
    def demo(self, signals='all', N=None, dft=None):
        """Plots signal waveforms, and optionally their DFTs.

        # Arguments:
            signals: str / [str] / [(str, dict)]
                'all' will set `signals = TestSignals.DEMO`, and plot in
                that order. Else, strings must be in `TestSignals.SUPPORTED`.
                Can also be `(str, dict)` pairs in a list, dict passed as
                keyword arguments to the generating function.
                Also see `help(ssqueezepy._test_signals)`, and
                `help(TestSignals.make_signals)`.

            N: int
                Length (# of samples) of generated signals.

            dft: None / str['rows', 'cols']
                If not None, will also plot DFT of each signal along the signal.
                If `'cols'`, will stack horizontally - if `'rows'`, vertically.
        """
        data = self.make_signals(signals, N)
        if dft not in (None, 'rows', 'cols'):
            raise ValueError(f"`dft` must be 'rows', 'cols', or None (got {dft})")
        elif dft == 'cols':
            dft_kw = dict(ncols=2, h=.55, w=1)
        elif dft == 'rows':
            dft_kw = dict(nrows=2)

        for name, (x, t, (fparams, aparams)) in data.items():
            title = self._title(name, len(x), fparams, aparams)
            if dft:
                axrf = np.abs(rfft(x))
                pkw = [{'title': title}, {'title': f"rDFT({name})"}]
                plots([t, None], [x, axrf], pkw=pkw, show=1, **dft_kw)
            else:
                plot(t, x, show=1, title=title)

    def test_transforms(self, fn, signals='all', N=None):
        """Make `fn` return `None` to skip visuals (e.g. if already done by `fn`).

        Input signature is `fn(x, t, params, ...)`, where
        `params = (name, fparams, aparams)`. Output, if not None, must be
        `(Tf, pkw)`, where `Tf` is a 2D np.ndarray time-frequency transform,
        and `pkw` is keyword arguments to `ssqueezepy.visuals.imshow`
        (can be empty dict).

        Also see `help(ssqueezepy._test_signals)`, and
        `help(TestSignals.make_signals)`.
        """
        data = self.make_signals(signals, N)
        default_pkw = dict(abs=1, cmap='jet', show=1)

        for name, (x, t, (fparams, aparams)) in data.items():
            out = fn(x, t, (name, fparams, aparams))

            if out is not None:
                out, pkw = out
                default_pkw['title'] = self._title(name, len(x), fparams, aparams)
                for k, v in default_pkw.items():
                    pkw[k] = pkw.get(k, v)

                if isinstance(out, (tuple, list)):
                    for o in out:
                        imshow(o, **pkw)
                else:
                    imshow(out, **pkw)

    #### utils ###############################################################
    def make_signals(self, signals='all', N=None):
        """Generates `signals` signals of length `N`.

        Returns dictionary of `{name: x, t, (fparams, aparams)}`, where `x` is
        the signal, `t` is its time vector, `fparams` is a dict of keyword args
        to the carrier, and `aparams` to the amplitude modulator (if applicable,
        e.g. `lchirp:am-sine').
        `fparams` may additionally contain a special kwarg: `snr`, not passed to
        carrier `fn`, that adds random normal noise of SNR `snr` to signal.

        Also see `help(ssqueezepy._test_signals)`.
        """
        def _process_args(name, fparams, aparams):
            fname, aname = (name.split(':') if ':' in name else
                            (name, ''))
            fname, aname = fname.replace('-', '_'), aname.replace('-', '_')
            fname = fname.lstrip('#')  # in case present

            fn  = (getattr(self, fname) if fname else
                   lambda *args, **kw: (np.ones(args[0]), None))
            afn = (getattr(self, aname) if aname else
                   lambda *args, **kw: (np.ones(args[0]), None))

            tkw = {}
            for dc in (fparams, aparams):  # `aparams` take precedence
                for k, v in dc.items():
                    if k in ('tmin', 'tmax', 'endpoint'):
                        tkw[k] = v
            return fn, afn, fname, aname, tkw

        names, params_all = self._process_input(signals)

        data = {}
        for name, (fparams, aparams) in zip(names, params_all):
            fn, afn, *_, tkw = _process_args(name, fparams, aparams)
            snr = fparams.pop('snr', self.snr)

            x, t = fn(N, **fparams)
            x *= afn(len(x), **aparams, **tkw)[0]

            if name[0] == '#':
                x += x[::-1]
            if snr:
                noise_var = x.var() / 10**(snr/10)
                if self.seed is not None:
                    np.random.seed(self.seed)
                noise = np.sqrt(noise_var) * np.random.randn(len(x))
                # use actual values
                fparams['snr'] = 10*np.log10(x.var() / noise.var())

                x += noise

            data[name] = (x, t, (fparams, aparams))
        return data

    @classmethod
    def _title(self, signal, N, fparams, aparams, x=None, wrap_len=70):
        fparams = self._process_varname_alias(signal, N, fparams)
        snr = fparams.pop('snr', None)

        if snr:
            snr = "{:.1f}dB".format(snr)
            fparams = dict(N=N, SNR=snr, **fparams)
        else:
            fparams = dict(N=N, **fparams)
        # drop `.0` from integer floats
        fparams = {k: (int(v) if (isinstance(v, float) and v.is_integer()) else v)
                   for k, v in fparams.items()}

        ptxt = ', '.join(f"{k}={v}" for k, v in fparams.items())
        title = "{} | {}".format(signal, ptxt)

        if aparams:
            atxt = ', '.join(f"{k}={v}" for k, v in aparams.items())
            title += ', %s' % atxt

        title = _textwrap(title, wrap_len)
        return title

    @staticmethod
    def _process_varname_alias(signal, N, fparams):
        fparams = fparams.copy()
        for k, v in fparams.items():
            if (k == 'fmax' and v is None and
                    any(s in signal for s in ('lchirp', 'echirp', 'hchirp'))):
                fparams['fmax'] = N / 2
        return fparams

    def _process_params(self, N, tkw, fn=None, fmin=None, fmax=None):
        tkw = tkw.copy()
        for k in self.default_tkw:
            tkw[k] = tkw.get(k, self.default_tkw[k])

        if N is None:
            tmin, tmax = tkw['tmin'], tkw['tmax']
            if any(var is None for var in (tmin, tmax, fmin, fmax)):
                N = self.N
            else:
                f_fn = lambda *args, **kw: fn(*args, **kw, get_w=True)[1]
                N = self._est_N_nonalias(f_fn, tmin, tmax, fmin, fmax)

        if fmax is None:
            fmax = N // 2

        t = _t(**tkw, N=N)
        tmin, tmax = tkw['tmin'], tkw['tmax']
        return ((t, tmin, tmax, fmax) if fn else
                (t, tmin, tmax))

    def _est_N_nonalias(self, f_fn, tmin, tmax, fmin, fmax):
        """Find smallest `N` (number of samples) such that signal generated
        from `tmin` to `tmax` will not alias.

        https://dsp.stackexchange.com/a/72942/50076

        max_phi_increment = fmax_fn * (t[1] - t[0])
        t[1] - t[0] = (tmax - tmin) / (N - 1)  [[endpoint=True]]
        max_phi_increment = pi
        fmax_fn * (tmax - tmin) / (N - 1) = pi
        1 + fmax_fn * (tmax - tmin) / pi = N
        """
        # sample sufficiently finely
        t = np.linspace(tmin, tmax, 50000, endpoint=True)
        fmax_fn = np.max(f_fn(t, tmin, tmax, fmin, fmax))

        min_nonalias_N = int(np.ceil(1 + fmax_fn*(tmax - tmin)/pi))
        return min_nonalias_N

    def _process_input(self, signals):
        """
        `signals`:
            - Ensure is string, or list/tuple of strings or of lists/tuples,
            each list/tuple being a (str, dict) or (str, (dict, dict)) pair.
            - Ensure each string is in `SUPPORTED`, and has an accompanying
            `params` pair (if not, set from `defalt_args`).
            - Loads parameters into two separate dictionaries, one for
            'carrier' / base function, other for (amplitude) 'modulator'.
            Defaults loaded according to precedence: `name:am-name` overrides
            `name` and `am-name`, but latter two are used if former isn't set.
        """
        def raise_type_error(signal):
            raise TypeError("all tuple or list elements of `signals` "
                            "must be (str, dict) or (str, (dict, dict)) pairs "
                            "(got (%s))" % ', '.join(
                                map(lambda s: type(s).__name__, signal)))

        if isinstance(signals, (str, tuple)):
            if signals != 'all':
                signals = [signals]
        elif not isinstance(signals, list):
            raise TypeError("`signals` must be string, list, or tuple "
                            "(got %s)" % type(signals))

        if isinstance(signals, list):
            for signal in signals:
                if isinstance(signal, str):
                    if ':' in signal:
                        fname, aname = signal.split(':')
                    else:
                        fname, aname = signal, ''
                    fname = fname.lstrip('#')

                    for name in (fname, aname):
                        if name != '' and name not in self.SUPPORTED:
                            raise ValueError(f"'{name}' is not supported; "
                                             "must be one of: "
                                             + ", ".join(self.SUPPORTED))
                elif isinstance(signal, (list, tuple)):
                    if not (isinstance(signal[0], str) and
                            isinstance(signal[1], (dict, list, tuple))):
                        raise_type_error(signal)
                    elif (isinstance(signal[1], (list, tuple)) and
                          not (isinstance(signal[1][0], dict) and
                               isinstance(signal[1][1], dict))):
                        raise_type_error(signal)
                else:
                    raise TypeError("all elements of `signals` must be string, "
                                    "or tuple or list of (string, dict) or "
                                    "(string, (dict, dict)) pairs "
                                    "(found %s)" % type(signal))

        if signals == 'all':
            signals = self.DEMO.copy()
        elif not isinstance(signals, (list, tuple)):
            signals = [signals]

        names, params_all = [], []
        for signal in signals:
            if isinstance(signal, (tuple, list)):
                name, params = signal
                if isinstance(params, (list, tuple)):
                    fparams, aparams = params
                else:
                    fparams, aparams = params, {}
            else:
                name, fparams, aparams = signal, {}, {}

            if name[0] == '#':
                add_reversed = True
                name = name[1:]
            else:
                add_reversed = False

            if 'am-' in name:
                if name.startswith('am-'):
                    if name.endswith(':'):
                        name = name.rstrip(':')
                    fname, aname = 'cosine', name
                    defaults = (self.default_args.get(fname, {}),
                                self.default_args.get(aname, {}))
                    name = fname + ':' + aname
                else:
                    defaults = self.default_args.get(name, {})
                    fname, aname = name.split(':')

                if isinstance(defaults, (list, tuple)):
                    fdefaults, adefaults = defaults
                elif isinstance(defaults, dict) and defaults != {}:
                    fdefaults, adefaults = defaults, {}
                else:
                    fdefaults, adefaults = self.default_args.get(fname, {}), {}

                if adefaults == {}:
                    adefaults = self.default_args.get(aname, {})

                for k, v in fdefaults.items():
                    fparams[k] = fparams.get(k, v)
                for k, v in adefaults.items():
                    aparams[k] = aparams.get(k, v)

                if name.startswith('am-'):
                    fdefaults, adefaults = adefaults, fdefaults
            else:
                for k, v in self.default_args.get(name, {}).items():
                    fparams[k] = fparams.get(k, v)

            if add_reversed:
                name = '#' + name
            names.append(name)
            params_all.append([fparams, aparams])

        # store latest result for debug purposes
        self._names = names
        self._params_all = params_all
        return names, params_all

    #### prebuilt test methods ##############################################
    def wavcomp(self, wavelets, signals='all', N=None, w=1.2, h=None,
                tight_kw=None):
        """Plots CWT & SSQ_CWT taken with `wavelets` wavelets side by side,
        vertically.
        """
        if not isinstance(wavelets, (list, tuple)):
            wavelets = [wavelets]
        wavs = []
        for wavelet in wavelets:
            wavs.append(Wavelet._init_if_not_isinstance(wavelet))

        fn = lambda x, t, params: self._wavcomp_fn(
            x, t, params, wavelets, w=w, h=h, tight_kw=tight_kw)
        self.test_transforms(fn, signals=signals, N=N)

    def _wavcomp_fn(self, x, t, params, wavelets, w=1.2, h=None, tight_kw=None):
        def _get_default_hspace():
            """Set dims based on maximum number of rows titles occupy."""
            title_nrows = []
            for wavelet in wavelets:
                name, fparams, aparams = params
                title1, title2 = self._title_cwt(wavelet, name, x,
                                                 fparams, aparams)
                title_nrows.extend([title1.count('\n'), title2.count('\n')])

            max_rows = max(title_nrows) + 1
            return (.13 + .05*(max_rows - 2)) * (.9 / h)

        h = h or .45 * len(wavelets)
        fig, axes = plt.subplots(len(wavelets), 2, figsize=(w * 12, h * 12))

        for i, wavelet in enumerate(wavelets):
            Tx, Wx, *_ = ssq_cwt(x, wavelet, t=t)
            Tx = np.flipud(Tx)

            name, fparams, aparams = params
            title1, title2 = self._title_cwt(wavelet, name, x, fparams, aparams)

            pkw = dict(abs=1, cmap='jet', ticks=0, fig=fig)
            imshow(Wx, **pkw, ax=axes[i, 0], show=0, title=title1)
            imshow(Tx, **pkw, ax=axes[i, 1], show=0, title=title2)

        tight_kw = tight_kw or {}
        default_hspace = _get_default_hspace()
        defaults = dict(left=0, right=1, bottom=0, top=1, wspace=.01,
                        hspace=default_hspace)
        for k, v in defaults.items():
            tight_kw[k] = v
        plt.subplots_adjust(**tight_kw)
        plt.show()

    def cwt_vs_stft(self, wavelet, window, signals='all', N=None,
                    win_len=None, n_fft=None, window_name=None, config_str='',
                    w=1.2, h=.9, tight_kw=None):
        """Plots CWT & SSQ_CWT, and STFT & SSQ_STFT of `signals` taken with
        `wavelet` and `window` along the rest of parameters.

        `window_name` & `config_str` are used to title STFT plots. `w` & `h`
        control plots' width & height. `tight_kw` is passed to
        `plt.subplots_adjust()`.
        """
        fn = lambda x, t, params: self._cwt_vs_stft_fn(
            x, t, params, wavelet, window, win_len, n_fft, window_name,
            config_str, w, h, tight_kw)
        self.test_transforms(fn, signals=signals, N=N)

    def _cwt_vs_stft_fn(self, x, t, params, wavelet, window, win_len=None,
                        n_fft=None, window_name=None, config_str='', w=1.2, h=.9,
                        tight_kw=None):
        def _get_default_hspace():
            """Set dims based on maximum number of rows titles occupy."""
            max_rows = 1 + max(g.count('\n') for g in (ctitle1, ctitle2,
                                                       stitle1, stitle2))
            return (.13 + .05*(max_rows - 2)) * (.9 / h)

        fs = 1 / (t[1] - t[0])
        Tsx, Sx, *_ = ssq_stft(x, window, n_fft=n_fft, win_len=win_len, fs=fs)
        Twx, Wx, *_ = ssq_cwt(x, wavelet, t=t)
        Twx, Tsx, Sx = np.flipud(Twx), np.flipud(Tsx), np.flipud(Sx)

        name, fparams, aparams = params
        ctitle1, ctitle2 = self._title_cwt( wavelet, name, x, fparams, aparams)
        stitle1, stitle2 = self._title_stft(window,  name, x, fparams, aparams,
                                            win_len, n_fft, window_name,
                                            config_str)

        fig, axes = plt.subplots(2, 2, figsize=(w * 12, h * 12))

        pkw = dict(abs=1, cmap='jet', ticks=0, fig=fig)
        imshow(Wx,  **pkw, ax=axes[0, 0], show=0, title=ctitle1)
        imshow(Twx, **pkw, ax=axes[0, 1], show=0, title=ctitle2)
        imshow(Sx,  **pkw, ax=axes[1, 0], show=0, title=stitle1)
        norm = ((0, np.abs(Tsx).mean()*300) if ("packed-poly" in name)
                else None)
        norm = ((0, np.abs(Tsx).mean()*200) if ("#par-lchirp" in name)
                else norm)
        imshow(Tsx, **pkw, ax=axes[1, 1], show=0, title=stitle2, norm=norm)

        tight_kw = tight_kw or {}
        default_hspace = _get_default_hspace()
        defaults = dict(left=0, right=1, bottom=0, top=1, wspace=.01,
                        hspace=default_hspace)
        for k, v in defaults.items():
            tight_kw[k] = v
        plt.subplots_adjust(**tight_kw)
        plt.show()

    @staticmethod
    def _title_cwt(wavelet, name, x, fparams, aparams, wrap_len=53):
        title = TestSignals._title(name, len(x), fparams, aparams)

        # special case: GMW
        wname = wavelet.name.replace(' L1', '').replace(' L2', '')

        twav = '%s wavelet | %s' % (wname, wavelet.config_str)
        ctitle1 = title + '\nabs(CWT) | ' + twav
        ctitle2 = 'abs(SSQ_CWT)'

        ctitle1 = _textwrap(ctitle1, wrap_len)
        return ctitle1, ctitle2

    @staticmethod
    def _title_stft(window, name, x, fparams, aparams, win_len=None, n_fft=None,
                    window_name='', config_str='', wrap_len=53):
        title = TestSignals._title(name, len(x), fparams, aparams)

        if win_len is not None:
            twin = "{} window | win_len={}, n_fft={}, {}".format(
                window_name, win_len, n_fft, config_str)
        else:
            twin = "{} window | n_fft={}, {}".format(window_name, n_fft,
                                                     config_str)
        stitle1 = title + '\nabs(STFT) | ' + twin
        stitle2 = 'abs(SSQ_STFT)'

        stitle1 = _textwrap(stitle1, wrap_len)
        return stitle1, stitle2

    def ridgecomp(self, signals='all', N=None, penalty=20, n_ridges=2, bw=None,
                  transform='cwt', w=1.2, h=.4, **transform_kw):
        """Plots extracted ridges from a CWT or STFT and them SSQ'd of `signals`,
        superimposed on the transform itself, passing in `transform_kw` to
        `ssq_cwt` or `ssq_stft`. `w` & `h` control plots' width & height.

        See `help(ridge_extraction.extract_ridges)`.
        """
        fn = lambda x, t, params: self._ridgecomp_fn(
            x, t, params, penalty, n_ridges, bw, transform,
            **transform_kw)
        self.test_transforms(fn, signals=signals, N=N)

    def _ridgecomp_fn(self, x, t, params, penalty=20, n_ridges=2, bw=None,
                      transform='cwt', w=1.2, h=.4, **transform_kw):
        transform_fn = ssq_cwt if transform == 'cwt' else ssq_stft
        Tfs, Tf, ssq_freqs, scales, *_ = transform_fn(x, t=t, **transform_kw)

        if bw is None:
            tf_bw, ssq_bw = 10, 2
        elif isinstance(bw, tuple):
            tf_bw, ssq_bw = bw
        else:
            tf_bw = ssq_bw = bw
        rkw = dict(penalty=penalty, n_ridges=n_ridges, transform=transform)
        ridges     = extract_ridges(Tf,  scales,    bw=tf_bw,  **rkw)
        ssq_ridges = extract_ridges(Tfs, ssq_freqs, bw=ssq_bw, **rkw)

        name, fparams, aparams = params
        if transform == 'cwt':
            Tf = np.flipud(Tf)
            ridges = len(Tf) - ridges
            title, title_s = "abs(CWT) w/ ridges", "abs(SSQ_CWT) w/ ridges"
        else:
            title, title_s = "abs(STFT) w/ ridges", "abs(SSQ_STFT) w/ ridges"
        tridge = "\npenalty={}, n_ridges={}, tf_bw={}, ssq_bw={}".format(
            penalty, n_ridges, tf_bw, ssq_bw)
        title += tridge
        tbase = self._title(name, len(x), fparams, aparams)
        title = tbase + '\n' + title

        _, axes = plt.subplots(1, 2, figsize=(w * 12, h * 12))

        pkw = dict(color='k', linestyle='--', ylims=(0, len(Tf)),
                   xlims=(0, Tf.shape[1]), ticks=0)
        plot(ridges,     ax=axes[0], **pkw)
        imshow(Tf,  abs=1, title=title,   ax=axes[0], show=0)
        plot(ssq_ridges, ax=axes[1], **pkw)
        imshow(Tfs, abs=1, title=title_s, ax=axes[1], show=0)

        tight_kw = dict(left=0, right=1, bottom=0, top=1, wspace=.01, hspace=0)
        plt.subplots_adjust(**tight_kw)
        plt.show()


def _t(tmin, tmax, N, endpoint=False):
    return np.linspace(tmin, tmax, N, endpoint=endpoint)
